import json
import sys
from typing import List, Optional

import click

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import HaJobStates
from anyscale.controllers.job_controller import JobController
import anyscale.job
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group("job", help="Interact with production jobs running on Anyscale.")
def job_cli() -> None:
    pass


@job_cli.command(name="submit",)
@click.argument("job-config-file", required=True)
@click.option("--name", "-n", required=False, default=None, help="Name of the job.")
@click.option("--description", required=False, default=None, help="Description of job.")
@click.option(
    "--follow",
    "-f",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to follow the log of the created job",
)
@click.option(
    "--wait",
    "-w",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to block this CLI command until the job succeeds (or reaches another terminal state)",
)
@click.argument(
    "entrypoint", required=False, nargs=-1,
)
def submit(  # noqa: PLR0913
    job_config_file: str,
    entrypoint: List[str],
    name: Optional[str],
    description: Optional[str],
    follow: Optional[bool],
    wait: Optional[bool],
) -> None:
    """ Submit a Production Job

    This function accepts 1 argument, a path to a YAML config file that defines this job.


    To wait for the Job to complete, use the --wait flag as follows.

    $ anyscale job submit job.yaml --wait

    If running from an Anyscale Workspace, this command also support an alternative syntax to specify the entrypoint inline:

    $ anyscale job submit -- python script.py

    The Cluster Environment and Cluster Config are inferred from the Workspace Configuration if running from a workspace.

    """

    job_controller = JobController()
    id = job_controller.submit(  # noqa: A001
        job_config_file,
        name,
        description,
        entrypoint=entrypoint,
        is_entrypoint_cmd="--" in sys.argv,
    )
    if follow:
        job_controller.logs(id, should_follow=follow)
    if wait:
        job_controller.wait(job_id=id)


@job_cli.command(name="list", help="Display information about existing jobs.")
@click.option("--name", "-n", required=False, default=None, help="Filter by job name.")
@click.option(
    "--job-id", "--id", required=False, default=None, help="Filter by job id."
)
@click.option(
    "--project-id", required=False, default=None, help="Filter by project id."
)
@click.option(
    "--include-all-users",
    is_flag=True,
    default=False,
    help="Include jobs not created by current user.",
)
@click.option(
    "--include-archived",
    is_flag=True,
    default=False,
    help=(
        "List archived jobs as well as unarchived jobs."
        "If not provided, defaults to listing only unarchived jobs."
    ),
)
@click.option(
    "--max-items",
    required=False,
    default=10,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list(  # noqa: A001, PLR0913
    name: Optional[str],
    job_id: Optional[str],
    project_id: Optional[str],
    include_all_users: bool,
    include_archived: bool,
    max_items: int,
) -> None:
    job_controller = JobController()
    job_controller.list(
        name=name,
        job_id=job_id,
        project_id=project_id,
        include_all_users=include_all_users,
        include_archived=include_archived,
        max_items=max_items,
    )


@job_cli.command(name="archive", help="Archive a job.")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
def archive(job_id: Optional[str], name: Optional[str]) -> None:
    job_controller = JobController()
    job_controller.archive(job_id=job_id, job_name=name)


@job_cli.command(name="terminate", help="Attempt to terminate a job asynchronously.")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
def terminate(job_id: Optional[str], name: Optional[str]) -> None:
    job_controller = JobController()
    job_controller.terminate(job_id=job_id, job_name=name)


@job_cli.command(name="logs", help="Print the logs of a job.")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--follow",
    "-f",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to follow the log.",
)
def logs(job_id: Optional[str], name: Optional[str], follow: bool = False,) -> None:
    job_controller = JobController()
    job_controller.logs(
        job_id=job_id, job_name=name, should_follow=follow,
    )


@job_cli.command(name="wait")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--state",
    "-s",
    required=False,
    default=HaJobStates.SUCCESS,
    help="The state to wait for this Job to enter",
)
@click.option(
    "--timeout",
    "-t",
    required=False,
    default=None,
    type=float,
    help="The timeout in seconds after which this command will exit.",
)
def wait(
    job_id: Optional[str],
    name: Optional[str],
    state: str = HaJobStates.SUCCESS,
    timeout=None,
) -> None:
    """Wait for a Job to enter a specific state (default: SUCCESS).

    To specify the Job by name, use the --name flag. To specify the Job by id, use the --job-id flag.

    If the Job reaches the target state, the command will exit successfully.

    If the Job reaches a terminal state other than the target state, the command will exit with an error.

    If the command reaches the timeout, the command will exit with an error.
    """
    state = state.upper()
    job_controller = JobController()
    job_id = job_controller.wait(
        job_id=job_id,
        job_name=name,
        is_service=False,
        target_state=state,
        timeout_secs=timeout,
    )


@job_cli.group("output", help="Interact with Production Job outputs.")
def output_cli() -> None:
    pass


@output_cli.command(name="get")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
def get_output(job_id: Optional[str], name: Optional[str]) -> None:
    """Get the output of a Production Job."""
    job_controller = JobController()
    job_controller.retrieve_output(job_name=name, job_id=job_id)


@output_cli.command(name="write")
@click.option(
    "--file",
    "-f",
    required=False,
    type=click.File("rb"),
    help="File to read the output from.",
)
@click.argument(
    "input_str", required=False, type=str,
)
def write_output(file, input_str) -> None:
    """Write the output of a Production Job to Anyscale.

    This command should be called from within a Production Job entrypoint. It will record the JSON output of your Job.
    If you specify neither a file or a string, this command will read from STDIN.

    Examples:

        anyscale job output write -f output.json

        anyscale job output write '{"output": 5}'

        cat output.json | anyscale job output write

    """
    data = None
    if input_str and file:
        raise click.ClickException("Cannot specify both --file and an input string.")
    if input_str is not None:
        data = json.loads(input_str)
    else:
        data = json.load(file or sys.stdin)
    anyscale.job.output(data)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedProductionServiceV2VersionAPIModel(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_at': 'datetime',
        'weight': 'int',
        'version': 'str',
        'ray_serve_config': 'object',
        'build_id': 'str',
        'compute_config_id': 'str',
        'production_job_ids': 'list[str]'
    }

    attribute_map = {
        'created_at': 'created_at',
        'weight': 'weight',
        'version': 'version',
        'ray_serve_config': 'ray_serve_config',
        'build_id': 'build_id',
        'compute_config_id': 'compute_config_id',
        'production_job_ids': 'production_job_ids'
    }

    def __init__(self, created_at=None, weight=None, version=None, ray_serve_config=None, build_id=None, compute_config_id=None, production_job_ids=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedProductionServiceV2VersionAPIModel - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._created_at = None
        self._weight = None
        self._version = None
        self._ray_serve_config = None
        self._build_id = None
        self._compute_config_id = None
        self._production_job_ids = None
        self.discriminator = None

        self.created_at = created_at
        self.weight = weight
        self.version = version
        self.ray_serve_config = ray_serve_config
        self.build_id = build_id
        self.compute_config_id = compute_config_id
        self.production_job_ids = production_job_ids

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501

        Time the version was created  # noqa: E501

        :return: The created_at of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedProductionServiceV2VersionAPIModel.

        Time the version was created  # noqa: E501

        :param created_at: The created_at of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def weight(self):
        """Gets the weight of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501

        The percentage of traffic sent to this version. This is a number between 0 and 100.  # noqa: E501

        :return: The weight of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :rtype: int
        """
        return self._weight

    @weight.setter
    def weight(self, weight):
        """Sets the weight of this DecoratedProductionServiceV2VersionAPIModel.

        The percentage of traffic sent to this version. This is a number between 0 and 100.  # noqa: E501

        :param weight: The weight of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and weight is None:  # noqa: E501
            raise ValueError("Invalid value for `weight`, must not be `None`")  # noqa: E501

        self._weight = weight

    @property
    def version(self):
        """Gets the version of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501

        The version string identifier for this version  # noqa: E501

        :return: The version of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this DecoratedProductionServiceV2VersionAPIModel.

        The version string identifier for this version  # noqa: E501

        :param version: The version of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and version is None:  # noqa: E501
            raise ValueError("Invalid value for `version`, must not be `None`")  # noqa: E501

        self._version = version

    @property
    def ray_serve_config(self):
        """Gets the ray_serve_config of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501


        :return: The ray_serve_config of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :rtype: object
        """
        return self._ray_serve_config

    @ray_serve_config.setter
    def ray_serve_config(self, ray_serve_config):
        """Sets the ray_serve_config of this DecoratedProductionServiceV2VersionAPIModel.


        :param ray_serve_config: The ray_serve_config of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :type: object
        """
        if self.local_vars_configuration.client_side_validation and ray_serve_config is None:  # noqa: E501
            raise ValueError("Invalid value for `ray_serve_config`, must not be `None`")  # noqa: E501

        self._ray_serve_config = ray_serve_config

    @property
    def build_id(self):
        """Gets the build_id of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501

        The id of the cluster env build. This id will determine the docker image your Service is run using.  # noqa: E501

        :return: The build_id of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._build_id

    @build_id.setter
    def build_id(self, build_id):
        """Sets the build_id of this DecoratedProductionServiceV2VersionAPIModel.

        The id of the cluster env build. This id will determine the docker image your Service is run using.  # noqa: E501

        :param build_id: The build_id of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and build_id is None:  # noqa: E501
            raise ValueError("Invalid value for `build_id`, must not be `None`")  # noqa: E501

        self._build_id = build_id

    @property
    def compute_config_id(self):
        """Gets the compute_config_id of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501

        The id of the compute configuration that you want to use. This id will specify the resources required for your Service.The compute template includes a `cloud_id` that cannot be updated.  # noqa: E501

        :return: The compute_config_id of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._compute_config_id

    @compute_config_id.setter
    def compute_config_id(self, compute_config_id):
        """Sets the compute_config_id of this DecoratedProductionServiceV2VersionAPIModel.

        The id of the compute configuration that you want to use. This id will specify the resources required for your Service.The compute template includes a `cloud_id` that cannot be updated.  # noqa: E501

        :param compute_config_id: The compute_config_id of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and compute_config_id is None:  # noqa: E501
            raise ValueError("Invalid value for `compute_config_id`, must not be `None`")  # noqa: E501

        self._compute_config_id = compute_config_id

    @property
    def production_job_ids(self):
        """Gets the production_job_ids of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501

        The list of production job ids associated with this service version.  # noqa: E501

        :return: The production_job_ids of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :rtype: list[str]
        """
        return self._production_job_ids

    @production_job_ids.setter
    def production_job_ids(self, production_job_ids):
        """Sets the production_job_ids of this DecoratedProductionServiceV2VersionAPIModel.

        The list of production job ids associated with this service version.  # noqa: E501

        :param production_job_ids: The production_job_ids of this DecoratedProductionServiceV2VersionAPIModel.  # noqa: E501
        :type: list[str]
        """
        if self.local_vars_configuration.client_side_validation and production_job_ids is None:  # noqa: E501
            raise ValueError("Invalid value for `production_job_ids`, must not be `None`")  # noqa: E501

        self._production_job_ids = production_job_ids

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedProductionServiceV2VersionAPIModel):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedProductionServiceV2VersionAPIModel):
            return True

        return self.to_dict() != other.to_dict()



from typing import List, Optional

import click

from anyscale.cli_logger import BlockLogger
from anyscale.commands.util import OptionPromptNull
from anyscale.controllers.cloud_controller import CloudController


log = BlockLogger()  # CLI Logger


@click.group(
    "cloud",
    short_help="Configure cloud provider authentication for Anyscale.",
    help="""Configure cloud provider authentication and setup
to allow Anyscale to launch instances in your account.""",
)
def cloud_cli() -> None:
    pass


@cloud_cli.command(name="delete", help="Delete a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Delete cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to delete. Alternative to cloud name.",
    required=False,
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
def cloud_delete(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str], yes: bool
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().delete_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id, skip_confirmation=yes
    )


@cloud_cli.command(
    name="set-default",
    help=(
        "Sets default cloud for your organization. This operation can only be performed "
        "by organization admins, and the default cloud must have organization level "
        "permissions."
    ),
)
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Set cloud as default by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to set as default. Alternative to cloud name.",
    required=False,
)
def cloud_set_default(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().set_default_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id
    )


@cloud_cli.command(name="setup", help="Set up a cloud provider.")
@click.option(
    "--provider",
    help="The cloud provider type.",
    required=True,
    prompt="Provider",
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
)
@click.option(
    "--region",
    cls=OptionPromptNull,
    help="Region to set up the credentials in.",
    required=True,
    prompt="Region",
    default_option="provider",
    default=lambda p: "us-west-2" if p == "aws" else "us-west1",
    show_default=True,
)
@click.option("--name", "-n", help="Name of the cloud.", required=True, prompt="Name")
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
@click.option(
    "--gce",
    is_flag=True,
    default=False,
    hidden=True,
    help="GCE is the VM stack on GCP.",
)
@click.option(
    "--gcp-folder-id",
    help="Folder ID where the 'Anyscale'  folder will be created.",
    required=False,
    type=int,
)
@click.option(
    "--vpc-peering-ip-range", help="IP range for VPC peering.", required=False
)
@click.option(
    "--vpc-peering-target-project-id",
    help="Project ID of the target VPC.",
    required=False,
)
@click.option(
    "--vpc-peering-target-vpc-id", help="VPC ID of the target VPC.", required=False
)
@click.option(
    "--anyscale-managed",
    is_flag=True,
    default=False,
    help="Let anyscale create all the resources.",
)
def setup_cloud(  # noqa: PLR0913
    provider: str,
    region: str,
    name: str,
    yes: bool,
    gce: bool,
    anyscale_managed: bool,
    gcp_folder_id: Optional[int],
    vpc_peering_ip_range: Optional[str],
    vpc_peering_target_project_id: Optional[str],
    vpc_peering_target_vpc_id: Optional[str],
) -> None:

    if provider == "aws":  # only provides Anyscale Managed option for AWS
        if (
            vpc_peering_ip_range
            or vpc_peering_target_vpc_id
            or vpc_peering_target_project_id
        ):
            raise click.ClickException(
                "VPC peering option is is only available for GCP."
            )
        CloudController().setup_managed_cloud(
            provider=provider,
            region=region,
            name=name,
            yes=yes,
            folder_id=gcp_folder_id,
        )
    elif provider == "gcp":
        if anyscale_managed:
            CloudController().setup_managed_cloud(
                provider=provider,
                region=region,
                name=name,
                yes=yes,
                folder_id=gcp_folder_id,
            )
        else:
            CloudController().setup_cloud(
                provider=provider,
                region=region,
                name=name,
                yes=yes,
                gce=gce,
                folder_id=gcp_folder_id,
                vpc_peering_ip_range=vpc_peering_ip_range,
                vpc_peering_target_project_id=vpc_peering_target_project_id,
                vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
            )


@cloud_cli.command(
    name="list", help=("List information about clouds in your Anyscale organization."),
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Name of cloud to get information about.",
)
@click.option(
    "--cloud-id",
    "--id",
    required=False,
    default=None,
    help=("Id of cloud to get information about."),
)
def list_cloud(name: Optional[str], cloud_id: Optional[str],) -> None:
    print(CloudController().list_clouds(cloud_name=name, cloud_id=cloud_id,))


@cloud_cli.group("config", help="Manage the configuration for a cloud.")
def cloud_config_group() -> None:
    pass


@cloud_config_group.command("update", help="Update the configuration for a cloud.")
@click.argument("cloud-name", required=False)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to update. Alternative to cloud name.",
    required=False,
)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--max-stopped-instances",
    help="Maximum number of stopped instances permitted in the shared instance pool.",
    required=True,
)
def cloud_config_update(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    max_stopped_instances: int,
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().update_cloud_config(
        cloud_name=cloud_name or name,
        cloud_id=cloud_id,
        max_stopped_instances=max_stopped_instances,
    )


@cloud_config_group.command("get", help="Get the current configuration for a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to get details about. Alternative to cloud name.",
    required=False,
)
def cloud_config_get(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    print(
        CloudController().get_cloud_config(
            cloud_name=cloud_name or name, cloud_id=cloud_id,
        )
    )


@cloud_cli.command(
    name="secrets",
    help="[Deprecated] Allow clusters started in this cloud to access secrets manager.",
    hidden=True,
)
@click.argument("cloud-name", required=False)
@click.option(
    "--name",
    "-n",
    help="Allow clusters started in this cloud (identified by cloud name) to access secrets manager.",
    type=str,
)
@click.option(
    "--cloud-id",
    "--id",
    help="Allow clusters started in this cloud (identified by cloud id) to access secrets manager.",
    required=False,
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
@click.option(
    "--write-permissions",
    is_flag=True,
    default=False,
    help="Allow clusters starting in this cloud to write to secrets manager as well.",
)
def cloud_secrets(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    write_permissions: bool,
    yes: bool,
) -> None:
    log.warning(
        "`anyscale cloud secrets` has been deprecated. Please instead use the cloud provider "
        "API to add the appropriate policies to access the secrets manager to your "
        "dataplane IAM role. We will remove this command at the end of Q1 2023."
    )
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().experimental_setup_secrets(
        cloud_name=cloud_name or name,
        cloud_id=cloud_id,
        write_permissions=write_permissions,
        yes=yes,
    )


@cloud_cli.command(
    name="register", help="Register an anyscale cloud with your own resources."
)
@click.option(
    "--provider",
    help="The cloud provider type.",
    required=True,
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
)
@click.option(
    "--region",
    cls=OptionPromptNull,
    help="Region to set up the credentials in.",
    required=True,
    default_option="provider",
    default=lambda p: "us-west-2" if p == "aws" else "us-west1",
    show_default=True,
)
@click.option(
    "--name", "-n", help="Name of the cloud.", required=True,
)
@click.option(
    "--vpc-id", help="The ID of the VPC.", required=False, type=str,
)
@click.option(
    "--subnet-ids",
    help="Comma separated list of subnet ids.",
    required=False,
    type=str,
)
@click.option(
    "--efs-id", help="The EFS ID.", required=False, type=str,
)
@click.option(
    "--anyscale-iam-role-id",
    help="The Anyscale IAM Role ARN.",
    required=False,
    type=str,
)
@click.option(
    "--instance-iam-role-id",
    help="The instance IAM role ARN.",
    required=False,
    type=str,
)
@click.option(
    "--security-group-ids",
    help="IDs of the security groups.",
    required=False,
    type=str,
)
@click.option(
    "--s3-bucket-id", help="S3 bucket ID.", required=False, type=str,
)
@click.option(
    "--private-network", help="Use private network.", is_flag=True, default=False,
)
def register_cloud(  # noqa: PLR0913
    provider: str,
    region: str,
    name: str,
    vpc_id: str,
    subnet_ids: str,
    efs_id: str,
    anyscale_iam_role_id: str,
    instance_iam_role_id: str,
    security_group_ids: str,
    s3_bucket_id: str,
    private_network: bool,
) -> None:

    if provider == "aws":
        missing_args: List[str] = []
        for resource in [
            (vpc_id, "--vpc-id"),
            (subnet_ids, "--subnet-ids"),
            (efs_id, "--efs-id"),
            (anyscale_iam_role_id, "--anyscale-iam-role-id"),
            (instance_iam_role_id, "--instance-iam-role-id"),
            (security_group_ids, "--security-group-ids"),
            (s3_bucket_id, "--s3-bucket-id"),
        ]:
            if resource[0] is None:
                missing_args.append(resource[1])

        if len(missing_args) > 0:
            raise click.ClickException(f"Please provide a value for {missing_args}")

        CloudController().register_aws_cloud(
            region=region,
            name=name,
            vpc_id=vpc_id,
            subnet_ids=subnet_ids.split(","),
            efs_id=efs_id,
            anyscale_iam_role_id=anyscale_iam_role_id,
            instance_iam_role_id=instance_iam_role_id,
            security_group_ids=security_group_ids.split(","),
            s3_bucket_id=s3_bucket_id,
            private_network=private_network,
        )
    elif provider == "gcp":
        # TODO (allenyin): implement this
        raise click.ClickException(
            "Registering cloud for GCP is not currently supported"
        )
    else:
        raise click.ClickException(
            f"Invalid Cloud provider: {provider}. Available providers are [aws, gcp]."
        )


@cloud_cli.command(name="verify", help="Checks the healthiness of a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Verify cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Verify cloud by cloud id, alternative to cloud name.",
    required=False,
)
def cloud_verify(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> bool:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )

    return CloudController().verify_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id
    )



"""
Commands to interact with the Anyscale API
"""

import click

from anyscale.commands.anyscale_api.session_commands_commands import session_commands
from anyscale.commands.anyscale_api.session_operations_commands import (
    session_operations_commands,
)
from anyscale.commands.anyscale_api.sessions_commands import sessions


@click.group(
    "api", hidden=True, help="Various commands to interact with the Anyscale API."
)
def anyscale_api() -> None:
    pass


anyscale_api.add_command(session_commands)
anyscale_api.add_command(sessions)
anyscale_api.add_command(session_operations_commands)



import json
import os

import click

from anyscale.authenticate import (
    CREDENTIALS_DIRS_PERMISSIONS,
    CREDENTIALS_ENVVAR,
    CREDENTIALS_FILE,
    CREDENTIALS_FILE_PERMISSIONS,
)
from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client.models.user_info import UserInfo
from anyscale.util import credentials_check_sanity, get_endpoint, get_user_info


class AuthController:
    """The controller for auth CLI command. This does not inherit BaseController
    since it does not involve the actual authentication and interaction with the API.
    """

    def __init__(
        self,
        warn_envvar: bool = True,
        path: str = CREDENTIALS_FILE,
        log_output: bool = True,
    ):
        self.log = LogsLogger(log_output=log_output)
        self.url = get_endpoint("/credentials")
        self.path = os.path.expanduser(path)
        if warn_envvar:
            self._warn_environment_variable()

    def set(self, token: str = None) -> None:
        # if token is provided, skip user interaction
        if token is None:
            # check if the file already exists
            if os.path.exists(self.path) and not click.confirm(
                f"You already have credentials saved in {CREDENTIALS_FILE}. Do you want to overwrite?"
            ):
                self.log.log("Canceled")
                return

            # get token
            self.log.log(
                f"Please copy your credentials from {self.url} and paste them below."
            )
            token = click.prompt("Enter your credentials", hide_input=True).strip()

        # sanity check
        if not credentials_check_sanity(token):
            raise click.ClickException("Invalid credentials")

        # save and done
        self._save_credentials(token)

    def show(self) -> None:
        result = get_user_info()
        block_label = "Identifying the user"
        self.log.open_block(block_label)
        if result is None:
            self.log.info("Not authenticated", block_label=block_label)
        else:
            self.log.info("Successfully authenticated as:", block_label=block_label)
            self.log.info(
                self._auth_info_to_formatted_string(result), block_label=block_label
            )

        self.log.close_block()

    def fix(self) -> None:
        if not os.path.exists(self.path):
            raise click.ClickException(
                f"Failed to fix the permissions of {CREDENTIALS_FILE}. "
                "Please check if you can access the file."
            )
        os.chmod(self.path, CREDENTIALS_FILE_PERMISSIONS)
        self.log.log(f"Successfully fixed the permissions of {CREDENTIALS_FILE}")

    def remove(self) -> None:
        # check if the file exists
        if not os.path.exists(self.path):
            raise click.ClickException("Credentials file does not exist.")

        if not click.confirm(f"Are you sure you want to remove {CREDENTIALS_FILE}?"):
            self.log.log("Canceled")
            return

        os.remove(self.path)
        self.log.log(f"Removed the credentials from {CREDENTIALS_FILE}.")

    def _auth_info_to_formatted_string(self, userinfo: UserInfo) -> str:
        """
        Try to construct a formatted output str from the retrieved user info.
        If exception happens, just stringfy and return.
        """
        left_pad = " " * 2

        try:
            s = f"{left_pad}{'user name:': <30}{userinfo.name}\n"
            s += f"{left_pad}{'user email:': <30}{userinfo.email}\n"
            s += f"{left_pad}{'user id:': <30}{userinfo.id}\n"

            # NOTE: Currently, we only support 1 organization per user, thus we're flattening the structure.
            # However, this will require change in the future as it can be confusing
            for org in userinfo.organizations:
                s += f"{left_pad}{'organization name:': <30}{org.name}\n"
                s += f"{left_pad}{'organization id:': <30}{org.id}\n"
            s += f"{left_pad}{'organization role:': <30}{userinfo.organization_permission_level}"
            return s
        except AttributeError:
            return userinfo.to_str()

    def _save_credentials(self, token: str) -> None:
        dirname = os.path.dirname(self.path)
        if not os.path.exists(dirname):
            os.makedirs(dirname, CREDENTIALS_DIRS_PERMISSIONS)
        with open(self.path, "w", CREDENTIALS_FILE_PERMISSIONS) as f:
            f.write(self._auth_token_get_json_string(token))
            f.flush()
        os.chmod(self.path, CREDENTIALS_FILE_PERMISSIONS)
        self.log.log(f"Your credentials have been saved to {CREDENTIALS_FILE}")

    def _warn_environment_variable(self) -> None:
        if os.environ.get(CREDENTIALS_ENVVAR) is not None:
            self.log.warning(
                f"You have {CREDENTIALS_ENVVAR} variable set, which will override the credentials file"
            )

    def _auth_token_get_json_string(self, credentials_str: str) -> str:
        return json.dumps({"cli_token": credentials_str})



import os
from typing import IO, Optional

import click
import yaml
from yaml.loader import SafeLoader

from anyscale.controllers.config_controller import ConfigController
from anyscale.util import generate_slug


@click.group(
    "config",
    short_help="Manage configurations on Anyscale.",
    help=(
        "[DEPRECATED] Manages configurations such as cluster environments and "
        "compute configs that help you run jobs and clusters on Anyscale."
    ),
    hidden=True,
)
def config_cli() -> None:
    pass


@config_cli.command(
    name="convert",
    help="[DEPRECATED] Converts a Ray cluster YAML into a cluster environment and compute config.",
    hidden=True,
)
@click.argument("cluster-yaml-file", type=click.File("rb"), required=True)
@click.option("--cloud", help="Cloud to create compute_config for.", required=False)
@click.option(
    "--ml/--no-ml",
    is_flag=True,
    default=False,
    prompt="Need ML base image?",
    help="Whether to generate a cluster-env that contains ray-ml dependencies",
)
@click.option(
    "--gpu/--no-gpu",
    is_flag=True,
    default=False,
    prompt="Need GPU base image?",
    help="Whether to generate a cluster-env that contains the required dependencies to utilize a gpu",
)
def config_convert(
    cluster_yaml_file: IO[bytes], cloud: Optional[str], ml: bool, gpu: bool
) -> None:
    cluster_yaml = yaml.load(cluster_yaml_file, Loader=SafeLoader)
    cluster_env, compute_config = ConfigController().convert_cluster_yaml(
        cloud, cluster_yaml, ml, gpu
    )
    with open("cluster-env.yaml", "w") as f:
        yaml.safe_dump(cluster_env.to_dict(), f)
    folder_name = os.path.basename(os.getcwd())
    configs_generated_name = f"{folder_name}-{generate_slug()}"
    with open("compute-config.yaml", "w") as f:
        yaml.safe_dump(compute_config.to_dict(), f)
    print(
        "\ncluster-env.yaml and compute-config.yaml generated. Modify the either configs if desired "
        "(note any warnings from above) and then run the following to save the configs to the server:\n\n"
        f"anyscale config upload-configs cluster-env.yaml compute-config.yaml --name {configs_generated_name}"
    )


@config_cli.command(
    name="upload-configs",
    help="[DEPRECATED] Uploads both the cluster-env and compute-config.",
    hidden=True,
)
@click.argument("cluster-env-yaml-file", type=click.File("rb"), required=True)
@click.argument("compute-config-yaml-file", type=click.File("rb"), required=True)
@click.option(
    "--name", help="Name for both configs", required=True,
)
def upload_configs(
    cluster_env_yaml_file: IO[bytes], compute_config_yaml_file: IO[bytes], name: str
) -> None:
    cluster_env_yaml = yaml.load(cluster_env_yaml_file, Loader=SafeLoader)
    compute_config = yaml.load(compute_config_yaml_file, Loader=SafeLoader)
    ConfigController().create_cluster_env(name, cluster_env_yaml)
    ConfigController().create_compute_config(name, compute_config, False)
    print(
        "Cluster environment and compute config successfully created! To use these configs, either:\n"
        "1. specify it as an argument in your ray address like so: "
        f'RAY_ADDRESS="anyscale://?cluster_env={name}:1&cluster_compute={name}"\n'
        "2. specify it as an argument in your ray.init like so: "
        f'ray.init(address="anyscale://", cluster_env={name}:1, cluster_compute={name})'
    )


@config_cli.command(
    name="create-cluster-env",
    help="[DEPRECATED] Creates a cluster env from a cluster env yaml.",
    hidden=True,
)
@click.argument("cluster-env-yaml-file", type=click.File("rb"), required=True)
@click.option(
    "--name", help="Name of the cluster environment", required=True,
)
def create_cluster_env(cluster_env_yaml_file: IO[bytes], name: str) -> None:
    cluster_env_yaml = yaml.load(cluster_env_yaml_file, Loader=SafeLoader)
    ConfigController().create_cluster_env(name, cluster_env_yaml)
    print(
        "Cluster environment successfully created! To use this environment, "
        f"specify it as an argument in your ray address using ?cluster_env={name}:1"
    )


@config_cli.command(
    name="create-compute-config",
    help="[DEPRECATED] Creates a compute config from a compute config yaml.",
    hidden=True,
)
@click.argument("compute-config-yaml-file", type=click.File("rb"), required=True)
@click.option(
    "--name", help="Name of the compute-config", required=True,
)
@click.option(
    "--anonymous",
    "-a",
    help="Whether the compute-config should show in the Configurations list",
    required=True,
    default=False,
)
def create_compute_config(
    compute_config_yaml_file: IO[bytes], name: str, anonymous: bool,
) -> None:
    """
    TODO(mattweber): This function is deprecated for create_cluster_compute
    under the cluster_compute group of CLI commands
    """

    compute_config = yaml.load(compute_config_yaml_file, Loader=SafeLoader)
    ConfigController().create_compute_config(name, compute_config, anonymous)
    print(
        "Compute config successfully created! To use this config, "
        f"specify it as an argument in your ray address using ?cluster_compute={name}"
    )


@config_cli.command(
    name="setup-dev-ray",
    help="Starts or updates a cluster with the given name and configuration by installing "
    "the specified local ray repository on all nodes. "
    "Specifically the working-dir directory will be synced to the nodes of "
    'the cluster and the "ray" directory within will be installed. '
    "The cluster, the local version of ray installed on this machine, "
    "and the local ray repository must all be at the same ray commit. "
    "You can check the installed ray commit by running: "
    """

    ipython

    import ray

    ray.__commit__
    """,
    hidden=True,
)
@click.argument("cluster-name", type=str, required=True, default=None)
@click.argument("working-dir", type=str, required=True, default=None)
@click.option(
    "--cluster-env",
    "-e",
    type=str,
    help="Name of the cluster-env",
    required=False,
    default=None,
)
@click.option(
    "--cluster-compute",
    "-c",
    type=str,
    help="Name of the cluster-compute",
    required=False,
    default=None,
)
@click.option(
    "--timeout",
    "-t",
    help="Timeout in seconds for running setup-dev on all nodes.",
    required=False,
    default=100,
)
def setup_dev_ray(
    working_dir: str,
    cluster_name: str,
    cluster_env: Optional[str],
    cluster_compute: Optional[str],
    timeout: Optional[float],
) -> None:
    ConfigController().setup_dev_ray(
        working_dir, cluster_name, cluster_env, cluster_compute, timeout
    )
    print("The local ray repository has been installed on all nodes in the cluster.")



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import base64
import re
from typing import Any, Dict
import unicodedata


def execution_log_name(session_command_id: str) -> str:
    return "/tmp/ray_command_output_{session_command_id}".format(
        session_command_id=session_command_id
    )


def startup_log_name(session_id: str) -> str:
    return f"/tmp/session_startup_logs_{session_id}"


def startup_log_name_v2(session_id: str, encoded: bool = False) -> str:
    startup_log_name = "/var/log/anyscale/session_startup_logs_{session_id}".format(
        session_id=session_id
    )

    if encoded:
        # The encoding should be the same as `encodePathToFileName` in go/infra/anyscaled/internal/logcache/utils.go
        return base64.standard_b64encode(f"{startup_log_name}.out".encode()).decode()
    else:
        return startup_log_name


def slugify(value: str) -> str:
    """
    Code adopted from here https://github.com/django/django/blob/master/django/utils/text.py

    Convert  to ASCII. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Also strip leading and trailing whitespace.
    """

    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value).strip()
    return re.sub(r"[-\s]+", "-", value)


def get_autoscaler_container_path(path: str) -> str:
    """
    When autoscaler is run in a separate container, all locations Docker bind-mounted
    from the host to directory /a/b/c of the user container are instead bind-mounted
    to /home/ray/autoscaler_mounts/a/b/c in the autoscaler container.

    Args:
        path: Location in user container.
    Returns:
        Location in the autoscaler container.
    """
    return f"/home/ray/autoscaler_mounts/{path}"


def get_container_name(cluster_config: Dict[str, Any]) -> str:
    return str(cluster_config.get("docker", {}).get("container_name", ""))



import json
import os
import platform
import stat
from typing import Dict, Optional, Tuple, Union

import click

import anyscale
from anyscale.api import (
    ApiClientWrapperExternal,
    ApiClientWrapperInternal,
    AsyncApiClientWrapperExternal,
    configure_open_api_client_headers,
    configure_tcp_keepalive,
    format_api_exception,
)
from anyscale.cli_logger import BlockLogger
from anyscale.client import openapi_client
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.rest import ApiException as ApiExceptionInternal
import anyscale.conf
from anyscale.sdk import anyscale_client
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as AnyscaleApi
import anyscale.shared_anyscale_utils.conf as shared_anyscale_conf


CREDENTIALS_FILE = "~/.anyscale/credentials.json"
CREDENTIALS_ENVVAR = "ANYSCALE_CLI_TOKEN"
CREDENTIALS_FILE_PERMISSIONS = 0o600
CREDENTIALS_DIRS_PERMISSIONS = 0o700


class AuthenticationBlock:
    """
    Class to perform authentication logic. This class should never be instantiated
    directly. Instead call `get_auth_api_client()` to get the cached AuthenticationBlock
    object.
    """

    def __init__(  # noqa: PLR0913
        self,
        cli_token: Optional[str] = None,
        host: Optional[str] = None,
        use_asyncio: bool = False,
        validate_credentials: bool = True,
        log_output: bool = True,
        raise_structured_exception: bool = False,
    ):
        """Create an AuthenticationBlock object with the following properties that are useful
        externally: host, credentials, api_client, anyscale_api_client.

        Arguments:
            cli_token (Optional[str]): CLI token that must be prefixed with "sss_". If
                passed, the credentials will not be loaded from the environment variable
                or credentials file.
            host (Optional[str]): Host address of Anyscale. Default host will be used if not
                provided.
            use_asyncio (bool): Use asyncio in anyscale external api client.
            validate_credentials (bool): If set to False, credentials that are
                passed will not be authenticated against the Anyscale host.
                `validate_credentials=False` should only be used when writing
                internal tests, because credentials and the host will not be
                authenticated against Anyscale if provided.
            log_output (bool): Whether actions should be logged.
            raise_structured_exception (bool): Whether to raise a structured
                exception (python class with fields), or a click exception with a human
                readable string.
        """
        self.log = BlockLogger(log_output=log_output)
        self.block_label = "Authenticating"
        self.log.open_block(self.block_label)
        self.host = host or shared_anyscale_conf.ANYSCALE_HOST
        if cli_token is not None:
            self.credentials = cli_token
            cli_token_location = "variable"
        else:
            (
                self.credentials,
                cli_token_location,
            ) = AuthenticationBlock._load_credentials()
        if validate_credentials:
            self._validate_credentials_format(self.credentials)

        self.api_client = self._instantiate_api_client(raise_structured_exception)
        self.anyscale_api_client = self._instantiate_anyscale_client(
            raise_structured_exception, use_asyncio=use_asyncio
        )
        if validate_credentials:
            self._validate_api_client_auth()
        anyscale.conf.CLI_TOKEN = self.credentials
        self.log.info(
            f"Loaded Anyscale authentication token from {cli_token_location}.",
            block_label=self.block_label,
        )
        self._warn_credential_file_permissions(cli_token_location)
        self.log.close_block(self.block_label)

    @property
    def product_api_client(self) -> DefaultApi:
        # TODO: shawnp. rename the field from api_client to product_api_client
        return self.api_client

    @property
    def base_api_client(self) -> AnyscaleApi:
        # TODO: shawnp. rename the field from anyscale_api_client to base_api_client
        return self.anyscale_api_client

    def _instantiate_api_client(self, raise_structured_exception: bool) -> DefaultApi:
        """
        Instantiates client to interact with our frontend APIs
        Args:
            raise_structured_exception (bool): Whether to raise a structured
                exception (python class with fields), or a click exception with a human
                readable string.
        """
        configuration = openapi_client.Configuration(host=self.host)
        configuration.connection_pool_maxsize = 100

        cookie = f"cli_token={self.credentials}"
        api_client = ApiClientWrapperInternal(
            configuration,
            cookie=cookie,
            raise_structured_exception=raise_structured_exception,
        )
        configure_open_api_client_headers(api_client, "CLI")
        configure_tcp_keepalive(api_client)
        api_instance = openapi_client.DefaultApi(api_client)
        return api_instance

    def _instantiate_anyscale_client(
        self, raise_structured_exception: bool, use_asyncio: bool = False,
    ) -> AnyscaleApi:
        """
        Instantiates client to interact with our externalized APIs

        Arguments:
            raise_structured_exception (bool): Whether to raise a structured
                exception (python class with fields), or a click exception with a human
                readable string.
            use_asyncio (bool): If this flag is set, the client will run requests in
                a threadpool. Invocations of the api client will return coroutines.
        """
        configuration = anyscale_client.Configuration(host=self.host + "/ext/v0")
        configuration.connection_pool_maxsize = 100

        api_client = (
            ApiClientWrapperExternal(
                configuration,
                cookie=f"cli_token={self.credentials}",
                raise_structured_exception=raise_structured_exception,
            )
            if not use_asyncio
            else AsyncApiClientWrapperExternal(
                configuration,
                cookie=f"cli_token={self.credentials}",
                raise_structured_exception=raise_structured_exception,
            )
        )
        configure_open_api_client_headers(api_client, "CLI")
        configure_tcp_keepalive(api_client)
        api_instance = anyscale_client.DefaultApi(api_client)
        return api_instance

    def _validate_credentials_format(self, credentials: str) -> None:
        # Existing token style
        if anyscale.util.credentials_check_sanity(credentials):
            return
        else:
            url = anyscale.util.get_endpoint("/credentials")
            raise click.ClickException(
                "Your user credentials are invalid. Please go to "
                f"{url} and follow the instructions to properly set your credentials."
            )

    def _validate_api_client_auth(self) -> None:
        """
        Authenticates credentials by calling /api/v2/userinfo. Credentials that
        are valid for the internal product API will also be valid for the external
        API.
        """
        old_raise_structured_exception_val = (
            self.api_client.api_client.raise_structured_exception
        )
        self.api_client.api_client.raise_structured_exception = True
        try:
            self.api_client.get_user_info_api_v2_userinfo_get()
        except ApiExceptionInternal as e:
            if e.status == 401:
                raise click.ClickException(
                    f"Your user credentials are invalid for {anyscale.util.get_endpoint('')}. Please go to "
                    f"{anyscale.util.get_endpoint('/credentials')} and follow the instructions to properly "
                    "set your credentials."
                )
            else:
                format_api_exception(e, "GET", "/api/v2/docs")
        except Exception as e:
            raise e
        finally:
            self.api_client.api_client.raise_structured_exception = (
                old_raise_structured_exception_val
            )

    def _warn_credential_file_permissions(self, filepath: str) -> None:
        """
        Check if the mode of the credentials file has non-zero permissions for group/others
        and warn user if the credential file is not user-only (i.e., last two octals should be 0o00).
        """
        path = os.path.expanduser(filepath)
        if not os.path.exists(path):
            return
        # Avoid false warnings in Windows, which does not use umask-based permissions
        if platform.system() == "Windows":
            return
        if stat.S_IMODE(os.stat(path).st_mode) & (stat.S_IRWXG | stat.S_IRWXO):
            self.log.warning(
                f"Permissions for your credentials file {CREDENTIALS_FILE} are not secure. "
                f"Run 'anyscale auth fix' to make your credentials secure and not accessible by others."
            )

    @staticmethod
    def _load_credentials() -> Tuple[str, str]:
        # The environment variable ANYSCALE_CLI_TOKEN can be used to
        # overwrite the credentials in ~/.anyscale/credentials.json
        env_token = os.environ.get(CREDENTIALS_ENVVAR)
        if env_token is not None:
            return env_token, CREDENTIALS_ENVVAR
        path = os.path.expanduser(CREDENTIALS_FILE)
        if not os.path.exists(path):
            url = anyscale.util.get_endpoint("/credentials")
            host = anyscale.util.get_endpoint("")
            raise click.ClickException(
                "Credentials not found. You need to create an account at {} "
                "and then go to {} and follow the instructions.".format(host, url)
            )

        with open(path) as f:
            try:
                credentials: Dict[str, str] = json.load(f)
            except json.JSONDecodeError:
                msg = (
                    "Unable to load user credentials.\n\nTip: Try creating your "
                    "user credentials again by going to {} and "
                    "following the instructions. If this does not work, "
                    "please contact Anyscale support.".format(
                        anyscale.util.get_endpoint("/credentials")
                    )
                )
                raise click.ClickException(msg)
        received_token = credentials.get("cli_token")
        if received_token is None:
            raise click.ClickException(
                "The credential file is not valid. Please regenerate it by following "
                "the instructions at {}".format(
                    anyscale.util.get_endpoint("/credentials")
                )
            )
        return received_token, CREDENTIALS_FILE


_auth_api_client: Optional[Union[Exception, AuthenticationBlock]] = None


def get_auth_api_client(  # noqa: PLR0913
    cli_token: Optional[str] = None,
    host: Optional[str] = None,
    use_asyncio: bool = False,
    validate_credentials: bool = True,
    log_output: bool = True,
    raise_structured_exception: bool = False,
) -> AuthenticationBlock:
    """
    Function to get global AuthenticationBlock object. The AuthenticationBlock will only be
    instantiated once with the passed parameters from the first call to get_auth_api_client
    and will be cached and returned on subsequent calls. If the first instantiation
    of AuthenticationBlock raises an error, that error will be saved and raised on
    subsequent calls to get_auth_api_client.
    """
    global _auth_api_client
    if isinstance(_auth_api_client, Exception):
        raise _auth_api_client
    if _auth_api_client is None:
        try:
            _auth_api_client = AuthenticationBlock(
                cli_token=cli_token,
                host=host,
                use_asyncio=use_asyncio,
                validate_credentials=validate_credentials,
                log_output=log_output,
                raise_structured_exception=raise_structured_exception,
            )
        except Exception as e:
            _auth_api_client = e
            raise e
    return _auth_api_client



import os
import time
from typing import Any, List, Optional

from packaging import version

from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import WandBRunDetails
from anyscale.sdk.anyscale_client import JobsQuery
from anyscale.shared_anyscale_utils.util import slugify
from anyscale.shared_anyscale_utils.utils.protected_string import ProtectedString
from anyscale.util import get_endpoint
from anyscale.utils.imports.gcp import try_import_gcp_secretmanager


"""anyscale/experimental_integrations.py: Experimental util functions for W&B integration and secret store prototypes."""

WANDB_API_KEY_NAME = "WANDB_API_KEY_NAME"  # pragma: allowlist secret
WANDB_PROJECT_NAME = "WANDB_PROJECT_NAME"
WANDB_GROUP_NAME = "WANDB_GROUP_NAME"

FLAG_WANDB_INTEGRATION_PROTOTYPE = "wandb-integration-prototype"

log = BlockLogger()  # Anyscale CLI Logger


def get_aws_secret(secret_name: str, **kwargs) -> ProtectedString:
    """
    Get secret value from AWS secrets manager.

    Arguments:
        secret_name: Key of your secret
        kwargs: Optional credentials passed in to authenticate instance
    """
    import boto3

    client = boto3.client("secretsmanager", **kwargs)
    response = client.get_secret_value(SecretId=secret_name)

    # Depending on whether the secret is a string or binary, one of these fields will be populated.
    if "SecretString" in response:
        secret = response.pop("SecretString")
    else:
        secret = response.pop("SecretBinary")

    return ProtectedString(secret)


def get_gcp_secret(secret_name: str, **kwargs) -> ProtectedString:
    """
    Get secret value from GCP secrets manager.

    Arguments:
        secret_name: Key of your secret
        kwargs: Optional credentials passed in to authenticate instance
    """
    import google

    secretmanager = try_import_gcp_secretmanager()

    client = secretmanager.SecretManagerServiceClient(**kwargs)
    _, project_name = google.auth.default()

    name = f"projects/{project_name}/secrets/{secret_name}/versions/latest"
    response = client.access_secret_version(request={"name": name})

    return ProtectedString(response.payload.data.decode())


def wandb_get_api_key() -> ProtectedString:
    """
    Returns W&B API key based on key set in WANDB_API_KEY_NAME in
    AWS or GCP secrets manager.

    Assumes instance is running with correct IAM role, so credentials
    don't have to be passed to access secrets manager.
    """
    secret_names: List[str] = []
    wandb_api_key_name_env_var = os.environ.get(WANDB_API_KEY_NAME)
    if wandb_api_key_name_env_var:
        secret_names.append(wandb_api_key_name_env_var)
    cluster_id = os.environ.get("ANYSCALE_SESSION_ID")
    api_client = get_auth_api_client(log_output=False).api_client

    # Get cloud from cluster to use the correct method to
    # get secrets from a cloud.
    if cluster_id:
        cluster = api_client.get_decorated_cluster_api_v2_decorated_sessions_cluster_id_get(
            cluster_id
        ).result
        cloud_id = cluster.cloud_id
        cloud = api_client.get_cloud_api_v2_clouds_cloud_id_get(cloud_id).result
    else:
        raise Exception(f"Unable to find cluster {cluster_id}")

    # Store alternate secret name to try fetching in list to ensure
    # backward compatibility with old secret naming convention
    if cloud.provider == "AWS":
        secret_names.append(f"anyscale_{cloud.id}/{cluster.creator_id}/wandb_api_key")
    elif cloud.provider == "GCP":
        # GCP secret names cannot include "/"
        secret_names.append(f"anyscale_{cloud.id}-{cluster.creator_id}-wandb_api_key")
    secret_names.append(f"wandb_api_key_{cluster.creator_id}")

    if cloud.provider != "AWS" and cloud.provider != "GCP":
        raise Exception(
            "The Anyscale W&B integration is currently only supported for AWS and GCP clouds."
        )
    for secret_name in secret_names:
        try:
            if cloud.provider == "AWS":
                region = cloud.region
                secret = get_aws_secret(secret_name, region_name=region)
            elif cloud.provider == "GCP":
                secret = get_gcp_secret(secret_name)
            return secret
        except Exception:  # noqa: BLE001
            log.info(f"Unable to fetch API key with name {secret_name}.")

    raise Exception("Unable to fetch API key from cloud secrets manager.")


def wandb_setup_api_key_hook() -> Optional[str]:
    """
    Returns W&B API key based on key set in WANDB_API_KEY_NAME in
    AWS or GCP secrets manager. This returns the API key in plain text,
    so take care to not save the output in any logs.

    The WANDB_SETUP_API_KEY_HOOK will point to this method so it will
    be called by the OSS WandbLoggerCallback. Because this is called
    before wandb.init(), any other setup can also be done here.
    """
    api_client = get_auth_api_client(log_output=False).api_client
    feature_flag_on = api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
        FLAG_WANDB_INTEGRATION_PROTOTYPE
    ).result.is_on
    if not feature_flag_on:
        return None

    protected_api_key = wandb_get_api_key()

    try:
        import ray

        # Check if ray <= 2.2.0 before setting project and group env vars
        # because OSS will directly call the set_wandb_project_group_env_vars
        # hook for ray > 2.2.0. Dev versions of ray are greater than all other
        # versions (eg: 3.0.0.dev0).
        if version.parse(ray.__version__) <= version.parse("2.2.0"):
            # Set environment variables to define default W&B project and group.
            set_wandb_project_group_env_vars()
    except ImportError:
        # Ray should always be installed on the head node, so we should
        # never reach this case.
        log.info(
            "Unable to import `ray` so setting default W&B project and group if "
            "these parameters are not already specified."
        )
        set_wandb_project_group_env_vars()

    # API key returned in plaintext because the OSS WandbLoggerCallback
    # accepts the API key as a string arguement.
    return protected_api_key._UNSAFE_DO_NOT_USE  # noqa: SLF001


def set_wandb_project_group_env_vars():
    """
    Set WANDB_PROJECT_NAME and WANDB_GROUP_NAME environment variables
    for the OSS WandbLoggerCallback to use, based on the default mapping
    for production jobs, workspaces, and Ray jobs.
    """
    api_client = get_auth_api_client(log_output=False).api_client
    feature_flag_on = api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
        FLAG_WANDB_INTEGRATION_PROTOTYPE
    ).result.is_on
    if not feature_flag_on:
        return

    wandb_project_default = None
    wandb_group_default = None
    if os.environ.get("ANYSCALE_HA_JOB_ID"):
        production_job_id = os.environ.get("ANYSCALE_HA_JOB_ID")
        production_job = api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            production_job_id
        ).result

        wandb_project_default = "anyscale_default_project"
        wandb_group_default = slugify(production_job.name)
    elif os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID"):
        workspace_id = os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID")
        workspace = api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get(
            workspace_id
        ).result

        wandb_project_default = slugify(workspace.name)
    elif _try_get_ray_job_id():
        ray_job_id = _try_get_ray_job_id()
        cluster_id = os.environ.get("ANYSCALE_SESSION_ID")
        if cluster_id:
            cluster_name = api_client.get_session_api_v2_sessions_session_id_get(
                cluster_id
            ).result.name
        else:
            cluster_name = "anyscale_default_project"
        wandb_project_default = slugify(cluster_name)
        wandb_group_default = slugify(f"ray_job_id_{ray_job_id}")

    if not os.environ.get(WANDB_PROJECT_NAME) and wandb_project_default:
        os.environ[WANDB_PROJECT_NAME] = wandb_project_default
    if not os.environ.get(WANDB_GROUP_NAME) and wandb_group_default:
        os.environ[WANDB_GROUP_NAME] = wandb_group_default


def wandb_send_run_info_hook(run: Any) -> None:
    """
    The WANDB_PROCESS_RUN_INFO points to this method and is called on
    the `run` output of `wandb.init()`.

    Send the W&B URL to the control plane, and populate the link back to
    Anyscale from the W&B run config.
    """
    auth_api_client = get_auth_api_client(log_output=False)
    api_client = auth_api_client.api_client
    anyscale_api_client = auth_api_client.anyscale_api_client

    feature_flag_on = api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
        FLAG_WANDB_INTEGRATION_PROTOTYPE
    ).result.is_on
    if not feature_flag_on:
        return

    try:
        import wandb
    except ImportError:
        raise Exception("Unable to import wandb.")

    assert isinstance(
        run, wandb.sdk.wandb_run.Run
    ), "`run` argument must be of type wandb.sdk.wandb_run.Run"

    if os.environ.get("ANYSCALE_HA_JOB_ID"):
        production_job_id = os.environ.get("ANYSCALE_HA_JOB_ID")
        production_job = api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            production_job_id
        ).result
        # Wait up to 5 sec to ensure the job run exists for the production job. We have previously
        # observed a delay in populating job runs in the DB when /api/snapshot has large
        # response in the old cluster snapshot service code. This should no longer
        # be an issue for the events stream stack.
        retry = 5
        while not production_job.last_job_run_id and retry > 0:
            time.sleep(1)
            production_job = api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
                production_job_id
            ).result
            retry -= 1
        if not production_job.last_job_run_id:
            log.info(
                "Unable to find latest job execution for this production Anyscale job."
            )
        api_client.put_production_job_wandb_run_details_api_v2_integrations_production_job_wandb_run_details_production_job_id_put(
            production_job_id=production_job_id,
            wand_b_run_details=WandBRunDetails(
                wandb_project_url=run.get_project_url(), wandb_group=run.group
            ),
        )
        run.config.anyscale_logs = get_endpoint(f"/jobs/{production_job_id}")
    elif os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID"):
        workspace_id = os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID")
        api_client.put_workspace_wandb_run_details_api_v2_integrations_workspace_wandb_run_details_workspace_id_put(
            workspace_id=workspace_id,
            wand_b_run_details=WandBRunDetails(
                wandb_project_url=run.get_project_url(), wandb_group=run.group
            ),
        )
        cluster_id = os.environ.get("ANYSCALE_SESSION_ID")
        if cluster_id:
            run.config.anyscale_logs = get_endpoint(
                f"/workspaces/{workspace_id}/{cluster_id}"
            )
    elif _try_get_ray_job_id():
        ray_job_id = _try_get_ray_job_id()
        cluster_id = os.environ.get("ANYSCALE_SESSION_ID")
        # TODO(nikita): Use ray.runtime_context.get_runtime_context().get_job_submission_id() instead of
        # API call. This will be available with Ray 2.3: https://github.com/ray-project/ray/issues/28089
        ray_job_resp = anyscale_api_client.search_jobs(
            JobsQuery(cluster_id=cluster_id, ray_job_id=ray_job_id)
        ).results
        # Wait up to 5 sec to send W&B url to control plane in case there
        # is delay populating ray job in DB.
        # We have previously observed this delay when /api/snapshot has large
        # response in the old cluster snapshot service code, but this should no longer
        # be an issue for the events stream stack.
        retry = 5
        while not len(ray_job_resp) and retry > 0:
            time.sleep(1)
            ray_job = anyscale_api_client.search_jobs(
                JobsQuery(cluster_id=cluster_id, ray_job_id=ray_job_id)
            ).results
            retry -= 1
        if len(ray_job_resp):
            ray_job = ray_job_resp[0]
            api_client.put_job_wandb_run_details_api_v2_integrations_job_wandb_run_details_job_id_put(
                job_id=ray_job.id,
                wand_b_run_details=WandBRunDetails(
                    wandb_project_url=run.get_project_url(), wandb_group=run.group
                ),
            )
            run.config.anyscale_logs = get_endpoint(
                f"/interactive-sessions/{ray_job.id}"
            )
        else:
            log.info("Unable to find Ray job in Anyscale to populate with W&B URL.")


def _try_get_ray_job_id():
    """
    Get the Ray job id using the Ray API.
    """
    try:
        import ray

        return ray.get_runtime_context().get_job_id()
    except Exception:  # noqa: BLE001
        # Alternatively get job id from environment variable
        # if error importing Ray. This should never happen because
        # this code is called from a hook in Ray.
        return os.environ.get("RAY_JOB_ID")



from datetime import datetime
import json
import os
from os import path
import shlex
import subprocess
from typing import Any, Dict, List, Optional, Sequence, Union

import click
import yaml

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import (
    AppConfig,
    Build,
    ComputeTemplateConfig,
    CreateComputeTemplate,
    Session,
    StopSessionOptions,
)
from anyscale.cluster_config import get_cluster_config
from anyscale.controllers.base_controller import BaseController
from anyscale.project import (
    get_and_validate_project_id,
    get_project_id,
    get_project_session,
    get_project_sessions,
    load_project_or_throw,
)
from anyscale.sdk.anyscale_client import StartSessionOptions, UpdateSession
from anyscale.sdk.anyscale_client.models.create_cluster import CreateCluster
from anyscale.shared_anyscale_utils.util import get_container_name, slugify
from anyscale.util import (
    canonicalize_remote_location,
    get_endpoint,
    get_working_dir,
    wait_for_session_start,
)
from anyscale.utils.aws_credentials_util import (
    get_credentials_as_env_vars_from_cluster_config,
)
from anyscale.utils.env_utils import set_env
from anyscale.utils.imports.all import try_import_ray


def get_head_node_ip(cluster_config: Union[Dict[str, Any], str]) -> Any:
    try_import_ray()
    from ray.autoscaler.sdk import get_head_node_ip

    return get_head_node_ip(cluster_config)


def rsync(*args: Any, **kwargs: Any) -> None:
    try_import_ray()
    from ray.autoscaler.sdk import rsync as ray_rsync

    ray_rsync(*args, **kwargs)


class SessionController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log

    def stop(
        self,
        session_name: Optional[str],
        delete: bool,
        workers_only: bool,
        keep_min_workers: bool,
    ) -> None:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        sessions = get_project_sessions(
            project_id, session_name, self.api_client, all_active_states=True
        )

        if not session_name and len(sessions) > 1:
            raise click.ClickException(
                "Multiple active sessions: {}\n"
                "Please specify the one you want to stop with --cluster-name.".format(
                    [session.name for session in sessions]
                )
            )

        for session in sessions:
            # Stop the session and mark it as stopped in the database.
            self.api_client.stop_session_api_v2_sessions_session_id_stop_post(
                session.id,
                StopSessionOptions(
                    terminate=True,
                    workers_only=workers_only,
                    keep_min_workers=keep_min_workers,
                    delete=delete,
                ),
            )

        session_names = [session.name for session in sessions]
        session_names_str = ", ".join(session_names)
        url = get_endpoint(f"/projects/{project_id}")
        self.log.info(
            f"Session {session_names_str} shutting down. View progress at {url}"
        )

    def start(  # noqa: PLR0913
        self,
        session_name: Optional[str],
        build_id: Optional[str],
        build_identifier: Optional[str],
        compute_config: str,
        idle_timeout: Optional[int],
        sync_files: bool,
    ) -> None:
        if sync_files:
            raise click.ClickException(
                "--sync-files has been deprecated and `anyscale start` will no longer ensure files "
                "synced to the working directory of the head node will also be synced on the workers."
            )
        message = (
            "Warning: `anyscale start` has been deprecated. Please try "
            'using `ray.init("anyscale://cluster_name")` for your use case and let the Anyscale team '
            "know if you are having any issues."
        )
        self.log.warning(message)

        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        session_name = self._get_or_generate_session_name(session_name, project_id)

        assert (
            build_id or build_identifier
        ), "--build-id or --build-identifier must be specified."
        if not build_id and build_identifier:
            build_id = self._get_build_id_from_build_identifier(
                build_identifier, project_id
            )
        assert build_id  # for mypy

        compute_template_id = self._register_compute_config(compute_config, project_id)

        session = self._create_or_update_session_data(
            session_name, project_id, build_id, compute_template_id, idle_timeout
        )
        self.anyscale_api_client.start_session(
            session.id,
            StartSessionOptions(
                build_id=build_id, compute_template_id=compute_template_id
            ),
        )

        wait_for_session_start(project_id, session_name, self.api_client)
        url = get_endpoint(f"/projects/{project_id}/clusters/{session.id}")
        print(f"Session {session_name} finished starting. View at {url}")

    def ssh(  # noqa: PLR0912, PLR0913, PLR0915
        self,
        session_name: str,
        ssh_option: Sequence[str],
        worker_node_id: Optional[str],
        worker_node_ip: Optional[str],
        force_head_node: bool,
        project_id: Optional[str],
        cmd: Optional[str] = None,
    ) -> None:

        config_path = "/home/ray/ray_bootstrap_config.yaml"
        on_head_node = path.exists(config_path)

        ssh_to_head = worker_node_id is None and worker_node_ip is None

        if force_head_node and ssh_to_head:
            raise click.ClickException(
                "Can't force head node if not sshing to a worker_node."
            )

        if worker_node_id is not None and worker_node_ip is not None:
            raise click.ClickException(
                "A worker node id and worker node ip cannot both be set at once."
            )

        if force_head_node or (on_head_node and (worker_node_id or worker_node_ip)):
            # On the head node
            print("SSHing from head node to worker node.")

            # Import ray where it's needed here to avoid making ray a
            # regular dependency of the CLI. This should only execute on
            # a head node where we can be confident ray will already
            # be installed.
            ray = try_import_ray()

            with open(config_path) as f:
                config_dict = yaml.safe_load(f)

            try:
                ssh_user = config_dict["auth"]["ssh_user"]
                ssh_key_path = config_dict["auth"]["ssh_private_key"]
                container_name = config_dict["docker"]["container_name"]
            except ValueError as e:
                raise click.ClickException(
                    f"Could not parse ~/ray_bootstrap_config.yaml: {e}"
                )

            if worker_node_ip is not None:
                node_ip = worker_node_ip
            elif worker_node_id is not None:
                ray.init(address="auto")
                nodes = ray.nodes()
                node_info = next(
                    (node for node in nodes if node.get("NodeID") == worker_node_id),
                    None,
                )
                if node_info is None:
                    raise click.ClickException(
                        f"Could not find a worker node with id {worker_node_id}."
                    )
                node_ip = node_info.get("NodeManagerAddress")

            ssh_command = (
                ["ssh"]
                + list(ssh_option)
                + ["-tt", "-i", ssh_key_path]
                + [f"{ssh_user}@{node_ip}"]
                + [f"docker exec -it {container_name} sh -c 'which bash && bash || sh'"]
            )
        else:
            # On a laptop
            project_id = get_and_validate_project_id(
                project_id=project_id,
                project_name=None,
                api_client=self.api_client,
                anyscale_api_client=self.anyscale_api_client,
            )
            session = get_project_session(project_id, session_name, self.api_client)

            cluster_config = get_cluster_config(
                session_name,
                self.api_client,
                disable_project_sync=True,
                project_id=project_id,
            )

            head_ip = self.api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get(
                session.id
            ).result.head_ip

            ssh_user = cluster_config["auth"]["ssh_user"]
            key_path = cluster_config["auth"]["ssh_private_key"]
            container_name = get_container_name(cluster_config)

            base_ssh_command = (
                ["ssh"]
                + list(ssh_option)
                + ["-tt", "-i", key_path]
                + [f"{ssh_user}@{head_ip}"]
            )
            if cmd:
                command = shlex.quote(cmd)
            else:
                command = shlex.quote("which bash && bash || sh")

            if ssh_to_head:
                ssh_command = base_ssh_command + (
                    [f"docker exec -it {container_name} sh -c {command}"]
                    if container_name
                    else []
                )
            else:
                if not container_name:
                    raise click.ClickException(
                        "Cannot ssh to a worker node if this cluster has no container."
                    )
                # Don't use the ssh_options for the second ssh command because
                # they probably refer to files that exist on the laptop but not
                # the head node.

                if worker_node_id is not None:
                    ssh_flag = f"--worker-node-id {worker_node_id}"
                else:
                    ssh_flag = f"--worker-node-ip {worker_node_ip}"
                ssh_from_head_command = (
                    f"'anyscale ssh {ssh_flag} --force-head-node True'"
                )
                ssh_command = base_ssh_command + (
                    [f"docker exec -it {container_name} sh -c  {ssh_from_head_command}"]
                )

        subprocess.run(ssh_command)  # noqa: B1

    def pull(  # noqa: PLR0913
        self,
        session_name: Optional[str] = None,
        source: Optional[str] = None,
        target: Optional[str] = None,
        config: Optional[str] = None,
        rsync_exclude_override: Optional[List[str]] = None,
    ) -> None:
        project_definition = load_project_or_throw()

        try:
            self.log.info("Collecting files from remote.")
            project_id = get_project_id(project_definition.root)
            cluster_config = get_cluster_config(session_name, self.api_client)
            if rsync_exclude_override is not None:
                cluster_config["rsync_exclude"] = rsync_exclude_override
            directory_name = get_working_dir(
                cluster_config, project_id, self.api_client
            )
            source_directory = f"{directory_name}/"

            aws_credentials = get_credentials_as_env_vars_from_cluster_config(
                cluster_config
            )

            source = canonicalize_remote_location(cluster_config, source, project_id)

            with set_env(**aws_credentials):
                if source and target:
                    target = os.path.expanduser(target)
                    rsync(
                        cluster_config, source=source, target=target, down=True,
                    )
                elif source or target:
                    raise click.ClickException(
                        "Source and target are not both specified. Please either specify both or neither."
                    )
                else:
                    rsync(
                        cluster_config,
                        source=source_directory,
                        target=project_definition.root,
                        down=True,
                    )

            if config:
                session = get_project_session(project_id, session_name, self.api_client)
                resp = self.api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get(
                    session.id
                )
                cluster_config = yaml.safe_load(resp.result.config_with_defaults)
                with open(config, "w") as f:
                    yaml.dump(cluster_config, f, default_flow_style=False)

            self.log.info("Pull completed.")

        except Exception as e:  # noqa: BLE001
            print(e)
            raise click.ClickException(str(e))  # type: ignore

    def push(  # noqa: PLR0913
        self,
        session_name: Optional[str],
        source: Optional[str],
        target: Optional[str],
        config: Optional[str],  # noqa: ARG002
        all_nodes: bool,
        rsync_exclude_override: Optional[List[str]] = None,
    ) -> None:
        if all_nodes:
            raise click.ClickException(
                "The `all_nodes` option is deprecated and will be removed in the future. "
                "Pushing to worker nodes is not reliable since workers may be "
                "added during autoscaling. Going forward, Anyscale will only support syncing "
                "working directory files to the head and workers."
            )

        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        session = get_project_session(project_id, session_name, self.api_client)
        session_name = session.name

        cluster_config = get_cluster_config(session_name, self.api_client)
        if rsync_exclude_override is not None:
            cluster_config["rsync_exclude"] = rsync_exclude_override
        target = canonicalize_remote_location(cluster_config, target, project_id)

        aws_credentials = get_credentials_as_env_vars_from_cluster_config(
            cluster_config
        )

        with set_env(**aws_credentials):
            if source and target:
                rsync(
                    cluster_config, source=source, target=target, down=False,
                )
            elif source or target:
                raise click.ClickException(
                    "Source and target are not both specified. Please either specify both or neither."
                )
            else:
                rsync(
                    cluster_config, source=None, target=None, down=False,
                )

        url = get_endpoint(f"/projects/{project_id}/clusters/{session.id}")
        self.log.info(f"Pushed to session {session_name}. View at {url}")

    # Helpers

    def _get_or_generate_session_name(
        self, session_name: Optional[str], project_id: Optional[str]
    ) -> str:
        """
        Return slugified session name if provided, else generate default session name from project id
        """
        assert (
            session_name or project_id
        ), "Either the session_name or project_id must be provided"

        if not session_name:
            session_name = str(
                self.api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get(
                    project_id=project_id,
                ).result.name
            )
        else:
            session_name = slugify(session_name)
        assert session_name
        return session_name

    def _register_compute_config(
        self, compute_config_file: str, project_id: str
    ) -> str:
        """
        Register the compute config with Anyscale and get its ID.
        """
        with open(compute_config_file) as f:
            config_dict = json.load(f)
        config_object = ComputeTemplateConfig(**config_dict)
        created_template = self.api_client.create_compute_template_api_v2_compute_templates_post(
            create_compute_template=CreateComputeTemplate(
                name=f"autogenerated-config-{datetime.now().isoformat()}",
                project_id=project_id,
                config=config_object,
            )
        ).result
        compute_template_id = str(created_template.id)
        return compute_template_id

    def _create_or_update_session_data(  # noqa: PLR0913
        self,
        session_name: str,
        project_id: str,
        build_id: str,
        compute_template_id: str,
        idle_timeout: Optional[int],
    ) -> Session:
        """
        Creates new session with app configs if session with `session_name` doesn't
        already exist. Otherwise, update the `idle_timeout` of the existing
        session if provided.
        """
        session_list = self.api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, active_only=False, name=session_name
        ).results

        session_exists = len(session_list) > 0
        if not session_exists:
            # Create a new session if there is no existing session with the givens session_name
            create_session_data = CreateCluster(
                name=session_name,
                project_id=project_id,
                cluster_environment_build_id=build_id,
                cluster_compute_id=compute_template_id,
                idle_timeout_minutes=idle_timeout,
            )
            session = self.anyscale_api_client.create_cluster(
                create_session_data
            ).result
        else:
            # Get the existing session and update the idle_timeout if required
            session = session_list[0]
            if idle_timeout:
                self.anyscale_api_client.update_session(
                    session.id, UpdateSession(idle_timeout=idle_timeout)
                )

        return session

    def _get_build_id_from_build_identifier(
        self, build_identifier: str, project_id: str
    ) -> str:
        # TODO(nikita): App config API should support getting a build ID from the build identifier.
        components = build_identifier.rsplit(":", 1)
        app_config_name = components[0]
        # try catch this
        try:
            app_config_revision = int(components[1]) if len(components) > 1 else None
        except ValueError:
            raise click.ClickException(
                "Invalid build-identifier provided. Please make sure the build identifier is of "
                "the form <app-config-name>:<build-revision>."
            )

        app_config_id = self._get_app_config_id_from_name(app_config_name, project_id)

        builds = self._list_builds(app_config_id)
        if app_config_revision:
            for build in builds:
                if build.revision == app_config_revision:
                    return str(build.id)

            raise click.ClickException(
                "Revision {} of app config '{}' not found.".format(
                    app_config_revision, app_config_name
                )
            )
        else:
            latest_build_revision = -1
            build_to_use = None
            for build in builds:
                if build.revision > latest_build_revision:
                    latest_build_revision = build.revision
                    build_to_use = build

            if not build_to_use:
                raise click.ClickException(
                    "Error finding latest build of app config {}. Please manually "
                    "specify the build version in the build identifier.".format(
                        app_config_name
                    )
                )
            return str(build_to_use.id)

    def _get_app_config_id_from_name(
        self, app_config_name: str, project_id: str
    ) -> str:
        app_configs = self._list_app_configs(project_id)
        for app_config in app_configs:
            if app_config.name == app_config_name:
                return str(app_config.id)

        raise click.ClickException(
            f"Application config '{app_config_name}' not found. "
            + "Available app configs: {}".format(", ".join(a.name for a in app_configs))
        )

    def _list_app_configs(self, project_id: str) -> List[AppConfig]:
        entities = []
        has_more = True
        while has_more:
            resp = self.anyscale_api_client.list_app_configs(
                project_id=project_id, count=50, paging_token=None
            )
            entities.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = paging_token is not None
        return entities

    def _list_builds(self, app_config_id: str) -> List[Build]:
        entities = []
        has_more = True
        while has_more:
            resp = self.anyscale_api_client.list_builds(
                app_config_id, count=50, paging_token=None
            )
            entities.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = paging_token is not None
        return entities

    def _resolve_session(
        self, session_name: str, project_name: Optional[str] = None
    ) -> Session:
        """
        Resolves a session by name.
        This is distinct from  `anyscale.project.get_project_session` because:
        1. we rely on project names instead of ids
        2. we allow non-active sessions to be resolved

        Raises an exception if the session does not exist.

        Params
        session_name - name of the session
        project_name - optional project name that the session is in;
                       if absent, we use the workspace's project
        """
        if project_name:
            projects = self.api_client.list_projects_api_v2_projects_get().results
            project = next(
                project for project in projects if project.name == project_name
            )
            project_id = project.id
        else:
            project_definition = load_project_or_throw()
            project_id = get_project_id(project_definition.root)

        sessions_list = self.api_client.list_sessions_api_v2_sessions_get(
            project_id, name=session_name
        ).results

        if len(sessions_list) == 0:
            raise click.ClickException(
                f"No session found with name {session_name} in project {project_id}"
            )

        return sessions_list[0]



from typing import Any, Dict, Optional, Union
from unittest.mock import ANY, Mock, patch

import click
from click.exceptions import ClickException
import pytest
import yaml
from yaml.loader import SafeLoader

from anyscale.conf import IDLE_TIMEOUT_DEFAULT_MINUTES
from anyscale.controllers.cluster_compute_controller import (
    ClusterComputeController,
    CreateClusterComputeConfigModel,
)
from anyscale.sdk.anyscale_client.models.archive_status import ArchiveStatus
from anyscale.sdk.anyscale_client.models.cluster_compute import ClusterCompute
from anyscale.sdk.anyscale_client.models.cluster_computes_query import (
    ClusterComputesQuery,
)
from anyscale.sdk.anyscale_client.models.clustercompute_list_response import (
    ClustercomputeListResponse,
)
from anyscale.sdk.anyscale_client.models.clustercompute_response import (
    ClustercomputeResponse,
)
from anyscale.sdk.anyscale_client.models.compute_template import ComputeTemplate
from anyscale.sdk.anyscale_client.models.compute_template_query import (
    ComputeTemplateQuery,
)
from anyscale.sdk.anyscale_client.models.create_cluster_compute import (
    CreateClusterCompute,
)
from anyscale.sdk.anyscale_client.models.text_query import TextQuery
from anyscale.utils.entity_arg_utils import EntityType, IdBasedEntity, NameBasedEntity


@pytest.fixture()
def mock_auth_api_client(mock_api_client: Mock, mock_sdk_client: Mock):
    mock_auth_api_client = Mock(
        api_client=mock_api_client, anyscale_api_client=mock_sdk_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.fixture()
def example_compute_config() -> Dict[str, Any]:

    example_compute_config_str = """

    {
    "cloud_id": "cld_HSrCZdMCYDe1NmMCJhYRgQ4p",
    "max_workers": 20,
    "maximum_uptime_minutes": null,
    "allowed_azs": null,
    "head_node_type": {
        "name": "head-node-type",
        "instance_type": "m5.2xlarge",
        "resources": null
    },
    "worker_node_types": [
        {
        "name": "worker-node-type-0",
        "instance_type": "m5.4xlarge",
        "resources": null,
        "min_workers": null,
        "max_workers": 10,
        "use_spot": false
        },
        {
        "name": "worker-node-type-1",
        "instance_type": "g4dn.4xlarge",
        "resources": null,
        "min_workers": null,
        "max_workers": 10,
        "use_spot": false
        }
    ],
    "aws_advanced_configurations_json": null,
    "gcp_advanced_configurations_json": null,
    "aws": null,
    "gcp": null,
    "azure": null
    }
    """
    return yaml.load(example_compute_config_str, Loader=SafeLoader)


# Used for generating the compute config read model from the write model.
# These fields may have these values if the customer didn't specify them,
# or because of backwards compatibility reasons.
def compute_config_with_default_fields(
    write_compute_config: Dict[str, Any]
) -> Dict[str, Any]:
    default_fields = {
        "idle_termination_minutes": IDLE_TIMEOUT_DEFAULT_MINUTES,
        "region": None,
    }
    return dict(**write_compute_config, **default_fields)


@pytest.fixture()
def mock_api_client() -> Mock:
    mock_api_client = Mock()

    return mock_api_client


@pytest.fixture()
def mock_sdk_client(cluster_compute_test_data: ClusterCompute) -> Mock:
    mock_sdk_client = Mock()

    mock_sdk_client.create_cluster_compute = Mock(
        return_value=ClustercomputeResponse(result=cluster_compute_test_data)
    )
    mock_sdk_client.delete_cluster_compute = Mock()
    mock_sdk_client.search_cluster_computes = Mock(
        return_value=ClustercomputeListResponse(results=[cluster_compute_test_data])
    )
    mock_sdk_client.get_cluster_compute = Mock()

    return mock_sdk_client


@pytest.mark.parametrize("cloud_id", [None, "test_cloud_id"])
@pytest.mark.parametrize("cloud", [None, "test_cloud_name"])
def test_create_validation(
    example_compute_config: Dict[str, Any],
    cloud_id: Optional[str],
    cloud: Optional[str],
):
    mock_get_cloud_id_and_name = Mock(return_value=("test_cloud_id", "test_cloud_name"))
    example_compute_config["cloud_id"] = cloud_id
    example_compute_config["cloud"] = cloud
    with patch.multiple(
        "anyscale.controllers.cluster_compute_controller",
        get_cloud_id_and_name=mock_get_cloud_id_and_name,
    ):
        if bool(cloud_id) + bool(cloud) != 1:
            with pytest.raises(click.ClickException):
                CreateClusterComputeConfigModel(**example_compute_config)
        else:
            CreateClusterComputeConfigModel(**example_compute_config)


@pytest.mark.parametrize("name", [None, "test_name"])
def test_create(
    mock_auth_api_client, name: Optional[str], example_compute_config: Dict[str, Any],
) -> None:
    cluster_compute_controller = ClusterComputeController()

    mock_load = Mock(return_value=example_compute_config)
    mock_cluster_compute_io = Mock()
    with patch.multiple("yaml", load=mock_load):
        cluster_compute_controller.create(mock_cluster_compute_io, name)

    cluster_compute_controller.anyscale_api_client.create_cluster_compute.assert_called_once_with(
        CreateClusterCompute(
            name=ANY, config=compute_config_with_default_fields(example_compute_config)
        )
    )


@pytest.mark.parametrize("cluster_compute_name", [None, "test_name"])
@pytest.mark.parametrize("id", [None, "cpt_123"])
def test_delete(
    mock_auth_api_client,
    cluster_compute_name: Optional[str],
    id: Optional[str],  # noqa: A002
    cluster_compute_test_data: ClusterCompute,
) -> None:
    cluster_compute_controller = ClusterComputeController()

    if cluster_compute_name is None and id is None:
        with pytest.raises(ClickException):
            cluster_compute_controller.delete(cluster_compute_name, id)
        cluster_compute_controller.anyscale_api_client.delete_cluster_compute.assert_not_called()
    elif cluster_compute_name is not None and id is not None:
        with pytest.raises(ClickException):
            cluster_compute_controller.delete(cluster_compute_name, id)
        cluster_compute_controller.anyscale_api_client.delete_cluster_compute.assert_not_called()
    elif cluster_compute_name and id is None:
        cluster_compute_controller.delete(cluster_compute_name, id)
        cluster_compute_controller.anyscale_api_client.search_cluster_computes.assert_called_once_with(
            ClusterComputesQuery(name=TextQuery(equals=cluster_compute_name))
        )
        cluster_compute_controller.anyscale_api_client.delete_cluster_compute.assert_called_once_with(
            cluster_compute_test_data.id
        )
    elif cluster_compute_name is None and id:
        cluster_compute_controller.delete(cluster_compute_name, id)
        cluster_compute_controller.anyscale_api_client.get_cluster_compute.assert_called_once_with(
            id
        )
        cluster_compute_controller.anyscale_api_client.delete_cluster_compute.assert_called_once_with(
            id
        )


@pytest.mark.parametrize(
    "cluster_compute_entity",
    [IdBasedEntity(id="cpt_123"), NameBasedEntity(name="compute_config_name")],
)
def test_archive(
    mock_auth_api_client,
    cluster_compute_entity: Union[IdBasedEntity, NameBasedEntity],
    compute_template_test_data: ComputeTemplate,
) -> None:
    cluster_compute_controller = ClusterComputeController()

    if cluster_compute_entity.type is EntityType.NAME:
        cluster_compute_controller.api_client.search_compute_templates_api_v2_compute_templates_search_post = Mock(
            return_value=Mock(results=[compute_template_test_data])
        )
        cluster_compute_controller.archive(cluster_compute_entity)
        cluster_compute_controller.api_client.search_compute_templates_api_v2_compute_templates_search_post.assert_called_once_with(
            ComputeTemplateQuery(
                orgwide=True,
                name={"equals": cluster_compute_entity.name},
                include_anonymous=True,
                archive_status=ArchiveStatus.ALL,
            )
        )
        cluster_compute_controller.api_client.archive_compute_template_api_v2_compute_templates_compute_template_id_archive_post.assert_called_once_with(
            compute_template_test_data.id
        )
    elif cluster_compute_entity.type is EntityType.ID:
        cluster_compute_controller.archive(cluster_compute_entity)
        cluster_compute_controller.anyscale_api_client.get_compute_template.assert_called_once_with(
            cluster_compute_entity.id
        )
        cluster_compute_controller.api_client.archive_compute_template_api_v2_compute_templates_compute_template_id_archive_post.assert_called_once_with(
            cluster_compute_entity.id
        )


@pytest.mark.parametrize("cluster_compute_name", [None, "test_cluster_compute_name"])
@pytest.mark.parametrize("cluster_compute_id", [None, "test_cluster_compute_id"])
def test_get(
    mock_auth_api_client,
    cluster_compute_name: Optional[str],
    cluster_compute_id: Optional[str],
    compute_template_test_data: ComputeTemplate,
):
    cluster_compute_controller = ClusterComputeController()
    if (cluster_compute_name is None and cluster_compute_id is None) or (
        cluster_compute_name is not None and cluster_compute_id is not None
    ):
        with pytest.raises(click.ClickException):
            cluster_compute_controller.get(
                cluster_compute_name=cluster_compute_name,
                cluster_compute_id=cluster_compute_id,
            )
            return
    else:
        with patch.multiple(
            "anyscale.controllers.cluster_compute_controller",
            get_cluster_compute_from_name=Mock(
                return_value=Mock(id=cluster_compute_id)
            ),
        ):
            cluster_compute_controller.api_client.get_compute_template_api_v2_compute_templates_template_id_get = Mock(
                return_value=Mock(result=compute_template_test_data)
            )
            cluster_compute_controller.get(
                cluster_compute_name=cluster_compute_name,
                cluster_compute_id=cluster_compute_id,
            )
            cluster_compute_controller.api_client.get_compute_template_api_v2_compute_templates_template_id_get.assert_called_once_with(
                cluster_compute_id
            )


@pytest.mark.parametrize("cluster_compute_name", [None, "test_cluster_compute_name"])
@pytest.mark.parametrize("cluster_compute_id", [None, "test_cluster_compute_id"])
@pytest.mark.parametrize("include_shared", [True, False])
def test_list(
    mock_auth_api_client,
    cluster_compute_name: Optional[str],
    cluster_compute_id: Optional[str],
    include_shared: bool,
):
    cluster_compute_controller = ClusterComputeController()
    cluster_compute_controller.anyscale_api_client.search_cluster_computes = Mock(
        return_value=Mock(results=[Mock()], metadata=Mock(next_paging_token=None))
    )
    cluster_compute_controller.list(
        cluster_compute_name=cluster_compute_name,
        cluster_compute_id=cluster_compute_id,
        include_shared=include_shared,
        max_items=50,
    )

    if cluster_compute_id:
        cluster_compute_controller.anyscale_api_client.get_cluster_compute.assert_called_once_with(
            cluster_compute_id
        )
    elif cluster_compute_name:
        cluster_compute_controller.anyscale_api_client.search_cluster_computes.assert_called_once_with(
            {"name": {"equals": cluster_compute_name}, "paging": {"count": 1}}
        )
    else:
        if not include_shared:
            cluster_compute_controller.api_client.get_user_info_api_v2_userinfo_get.assert_called_once_with()
            cluster_compute_controller.anyscale_api_client.search_cluster_computes.assert_called_once_with(
                {"creator_id": ANY, "paging": {"count": 20}}
            )



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
from typing import Union


def ray_semver_compare(a: str, b: str) -> int:
    """
    A dumb implementation of semver comparison because ray versions
    don't actually follow the semver standard.
    """
    a_parts = a.split(".")
    b_parts = b.split(".")
    for i, a_part in enumerate(a_parts):
        if i > len(b_parts):
            # No value in part is always less than any string in a part
            return 1

        try:
            parsed_a_part: Union[str, int] = int(a_part)
            parsed_b_part: Union[str, int] = int(b_parts[i])
        except ValueError:
            parsed_a_part = a_part
            parsed_b_part = b_parts[i]

        if parsed_a_part > parsed_b_part:  # type: ignore
            return 1
        elif parsed_a_part < parsed_b_part:  # type: ignore
            return -1
    if len(b_parts) > len(a_parts):
        # B still has parts, it must be bigger
        return -1

    return 0



import os
import shlex
import subprocess
from typing import Any, Dict, List, Optional, Sequence

from pydantic import BaseModel

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.experimental_workspace import (
    ExperimentalWorkspace,
)
from anyscale.cluster_config import configure_for_session
from anyscale.controllers.base_controller import BaseController
from anyscale.sdk.anyscale_client.models.cloud_version import CloudVersion
from anyscale.shared_anyscale_utils.conf import V2_RAY_MOUNT_PATH
from anyscale.util import get_working_dir
from anyscale.utils.imports.all import try_import_ray
from anyscale.workspace import load_workspace_id_or_throw, write_workspace_id_to_disk


class WorkspaceCommandContext(BaseModel):
    cluster_config: Dict[str, Any]
    head_ip: str
    project_id: str
    cluster_id: str


class WorkspaceController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log

    def clone(self, workspace: ExperimentalWorkspace) -> None:
        dir_name = workspace.name
        os.makedirs(dir_name)

        workspace_id = workspace.id
        write_workspace_id_to_disk(workspace_id, dir_name)

    def get_activated_workspace(self) -> ExperimentalWorkspace:
        workspace_id = load_workspace_id_or_throw()
        return self.api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get(
            workspace_id
        ).result

    def get_workspace_dir_name(self) -> str:
        """
        Currently the directory name of Workspaces/Clusters
        is determiend by the project name.
        """
        workspace = self.get_activated_workspace()
        project = self.api_client.get_project_api_v2_projects_project_id_get(
            workspace.project_id
        ).result
        return project.name

    def _get_workspace_command_context(self) -> WorkspaceCommandContext:
        workspace = self.get_activated_workspace()

        cluster_id = workspace.cluster_id
        project_id = workspace.project_id

        cluster_config = configure_for_session(
            cluster_id, api_client=self.api_client, disable_project_sync=True
        )
        head_ip = self.api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get(
            cluster_id
        ).result.head_ip

        return WorkspaceCommandContext(
            cluster_config=cluster_config,
            head_ip=head_ip,
            project_id=project_id,
            cluster_id=cluster_id,
        )

    def run_cmd(self, ssh_option: Sequence[str], cmd: str):
        workspace_command_context = self._get_workspace_command_context()
        cluster_config = workspace_command_context.cluster_config

        key_path = cluster_config["auth"]["ssh_private_key"]

        head_ip = workspace_command_context.head_ip
        ssh_user = cluster_config["auth"]["ssh_user"]
        target_host = f"{ssh_user}@{head_ip}"

        command = shlex.quote(cmd)

        ssh_command = (
            ["ssh"]
            + list(ssh_option)
            + ["-o", f"ProxyCommand=ssh -W %h:%p -i {key_path} {target_host}"]
            + ["-tt", "-i", key_path]
            + ["ray@0.0.0.0", "-p", "5020"]
            + [f"sh -c {command}"]
        )

        subprocess.run(ssh_command)  # noqa: B1

    def run_rsync(
        self,
        ssh_option: Sequence[str],
        local_path: str,
        *,
        down: bool,
        rsync_filters: List[str],
        rsync_excludes: List[str],
    ) -> None:
        try_import_ray()
        from ray.autoscaler.sdk import get_docker_host_mount_location

        workspace_command_context = self._get_workspace_command_context()
        cluster_config = workspace_command_context.cluster_config
        head_ip = workspace_command_context.head_ip
        project_id = workspace_command_context.project_id
        cluster_id = workspace_command_context.cluster_id

        ssh_user = cluster_config["auth"]["ssh_user"]
        key_path = cluster_config["auth"]["ssh_private_key"]

        base_ssh_command = ["ssh"] + ["-i", key_path] + list(ssh_option)

        directory_name = get_working_dir(cluster_config, project_id, self.api_client)

        cloud_id = self.anyscale_api_client.get_session(cluster_id).result.cloud_id
        cloud_version = self.anyscale_api_client.get_cloud(cloud_id).result.version
        mount_path = (
            V2_RAY_MOUNT_PATH
            if not cloud_version or cloud_version == CloudVersion.V2
            else get_docker_host_mount_location(cluster_config["cluster_name"])
        )
        source_directory = mount_path + directory_name + "/"

        rsync_command = [
            "rsync",
            "--rsh",
            subprocess.list2cmdline(base_ssh_command),
            "-avz",
            "--delete",
        ]

        for rsync_exclude in rsync_excludes:
            rsync_command.extend(["--exclude", rsync_exclude])

        for rsync_filter in rsync_filters:
            rsync_command.extend(["--filter", f"dir-merge,- {rsync_filter}"])

        if down:
            rsync_command += [
                f"{ssh_user}@{head_ip}:{source_directory}",
                local_path,
            ]
        else:
            rsync_command += [
                local_path,
                f"{ssh_user}@{head_ip}:{source_directory}",
            ]

        subprocess.run(rsync_command)  # noqa: B1



import json
import os
from typing import Optional
from unittest.mock import Mock, mock_open, patch

import click
import pytest

from anyscale.client.openapi_client.models.apply_production_service_v2_api_model import (
    ApplyProductionServiceV2APIModel,
)
from anyscale.client.openapi_client.models.create_production_service import (
    CreateProductionService,
)
from anyscale.client.openapi_client.models.production_job_config import (
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.serviceinternalmodel_list_response import (
    ServiceinternalmodelListResponse,
)
from anyscale.controllers.service_controller import ServiceConfig, ServiceController


TEST_SERVICE_ID = "test-service-id"
TEST_SERVICE_NAME = "test-service-name"


def test_service_config_model_with_entrypoint():
    config_dict = {
        "entrypoint": "python test.py",
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
        "healthcheck_url": "test.url",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.ray_serve_config is None


def test_service_config_model_with_ray_serve_config():
    config_dict = {
        "ray_serve_config": {"runtime_env": {"pip": ["requests"]}},
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.entrypoint is None


@pytest.mark.parametrize("use_default_project", [True, False])
@pytest.mark.parametrize("access", ["public", "private"])
@pytest.mark.parametrize("maximum_uptime_minutes", [60, None])
def test_update_service(
    mock_auth_api_client,
    use_default_project: bool,
    access: str,
    maximum_uptime_minutes: Optional[int],
) -> None:
    config_dict = {
        "entrypoint": "mock_entrypoint",
        "build_id": "mock_build_id",
        "compute_config_id": "mock_compute_config_id",
        "healthcheck_url": "mock_healthcheck_url",
        "access": access,
    }
    mock_logger = Mock()
    service_controller = ServiceController(log=mock_logger)
    if use_default_project:
        mock_infer_project_id = Mock(return_value="mock_default_project_id")
    else:
        mock_infer_project_id = Mock(return_value="mock_project_id")

    mock_validate_successful_build = Mock()

    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple(
        "anyscale.controllers.service_controller",
        infer_project_id=mock_infer_project_id,
    ), patch.object(
        ServiceController,
        "_get_maximum_uptime_minutes",
        return_value=maximum_uptime_minutes,
    ), patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ), patch.multiple(
        "os.path", exists=Mock(return_value=True)
    ):
        service_controller.deploy(
            "mock_config_file", name="mock_name", description="mock_description",
        )

    service_controller.api_client.apply_service_api_v2_decorated_ha_jobs_apply_service_put.assert_called_once_with(
        CreateProductionService(
            name="mock_name",
            description="mock_description",
            project_id="mock_default_project_id"
            if use_default_project
            else "mock_project_id",
            config=ProductionJobConfig(  # noqa: PIE804
                **{
                    "entrypoint": "mock_entrypoint",
                    "build_id": "mock_build_id",
                    "compute_config_id": "mock_compute_config_id",
                }
            ),
            healthcheck_url="mock_healthcheck_url",
            access=access,
        )
    )
    assert mock_logger.info.call_count == 4


def test_service_submit_parse_logic(mock_auth_api_client) -> None:
    service_controller = ServiceController()
    service_controller.generate_config_from_entrypoint = Mock()  # type: ignore
    service_controller.generate_config_from_file = Mock()  # type: ignore
    service_controller.deploy_from_config = Mock()  # type: ignore

    # We are not in a workspace, so entrypoint should not be allowed
    with pytest.raises(click.ClickException):
        service_controller.deploy(
            "file", None, None, entrypoint=["entrypoint"], is_entrypoint_cmd=False
        )

    with pytest.raises(click.ClickException):
        service_controller.deploy(
            "file", None, None, entrypoint=["entrypoint"], is_entrypoint_cmd=True
        )

    with pytest.raises(click.ClickException):
        service_controller.deploy(
            "file",
            None,
            None,
            entrypoint=["entrypoint", "commands"],
            is_entrypoint_cmd=True,
        )

    # Simulate a workspace
    with patch.dict(
        os.environ, {"ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": "fake_workspace_id"}
    ), patch.multiple(
        "anyscale.controllers.service_controller",
        override_runtime_env_for_local_working_dir=Mock(
            return_value={"rewritten": True}
        ),
    ):
        # Fails due to is_entrypoint_cmd being False
        with pytest.raises(click.ClickException):
            service_controller.deploy(
                "file", None, None, entrypoint=["entrypoint"], is_entrypoint_cmd=False
            )

        mock_config = Mock()
        service_controller.generate_config_from_file.return_value = mock_config
        service_controller.deploy(
            "file", None, None, entrypoint=[], is_entrypoint_cmd=False
        )
        service_controller.generate_config_from_file.assert_called_once_with(
            "file", None, None, healthcheck_url=None
        )
        service_controller.deploy_from_config.assert_called_once_with(mock_config)
        service_controller.generate_config_from_file.reset_mock()
        service_controller.deploy_from_config.reset_mock()

        mock_config = Mock()
        service_controller.generate_config_from_entrypoint.return_value = mock_config
        service_controller.deploy(
            "file", None, None, entrypoint=["entrypoint"], is_entrypoint_cmd=True
        )
        service_controller.generate_config_from_entrypoint.assert_called_once_with(
            ["file", "entrypoint"], None, None, healthcheck_url=None
        )
        service_controller.deploy_from_config.assert_called_once_with(mock_config)


@pytest.mark.parametrize("override_name", [None, "override_service_name"])
@pytest.mark.parametrize("override_version", [None, "override_version"])
@pytest.mark.parametrize("override_canary_weight", [None, 40])
def test_service_rollout(
    mock_auth_api_client,
    override_name: Optional[str],
    override_version: Optional[str],
    override_canary_weight: Optional[int],
) -> None:
    """
    This tests that when we submit a Service config with ray_serve_config,
    we submit both the Service v1 and Service v2 configs.
    Please reference services_internal_api_models.py for more details.
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    ray_serve_config = {"runtime_env": {"pip": ["requests"]}}
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"

    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_weight": 100,
    }
    mock_validate_successful_build = Mock()
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_controller.rollout(
            "test_ser8vice_config_file",
            name=override_name,
            version=override_version,
            canary_percent=override_canary_weight,
        )

    service_v2_config = ApplyProductionServiceV2APIModel(
        name=override_name if override_name else name,
        description=description,
        build_id=build_id,
        compute_config_id=compute_config_id,
        project_id=project_id,
        ray_serve_config=ray_serve_config,
        version=override_version if override_version else "test_abc",
        canary_weight=override_canary_weight if override_canary_weight else 100,
    )

    service_controller.api_client.apply_service_v2_api_v2_services_v2_apply_put.assert_called_once_with(
        service_v2_config
    )


class TestGetServiceID:
    def test_invalid_inputs(self, mock_auth_api_client):
        service_controller = ServiceController()

        with pytest.raises(
            click.ClickException,
            match="Service ID, name, or config file must be specified.",
        ):
            service_controller.get_service_id(None, None, None)

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id("test-id", "test-name", None)

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id("test-id", None, "test-config-file-path")

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                "test-id", "test-name", "test-config-file-path"
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                None, "test-name", "test-config-file-path"
            )

        with pytest.raises(click.ClickException, match="not found"):
            service_controller.get_service_id(None, None, "/fake/service/config/path")

    def test_service_id_provided(self, mock_auth_api_client):
        """If the ID is provided, the same ID should always be returned and no API calls made."""
        service_controller = ServiceController()

        assert service_controller.get_service_id("test-id", None, None) == "test-id"
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_not_called()

    def test_name_provided_no_matching_service(self, mock_auth_api_client):
        """If a name is provided but no services match it, an exception should be raised."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = ServiceinternalmodelListResponse(
            results=[]
        )

        with pytest.raises(
            click.ClickException,
            match=f"No service with name '{TEST_SERVICE_NAME}' was found",
        ):
            service_controller.get_service_id(None, TEST_SERVICE_NAME, None)

        # Test the config file codepath.
        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        with pytest.raises(
            click.ClickException,
            match=f"No service with name '{TEST_SERVICE_NAME}' was found",
        ):
            service_controller.get_service_id(None, None, "test-config-file")

    def test_name_provided_single_matching_service(self, mock_auth_api_client):
        """If a name is provided and one service matches it, its ID should be returned."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = ServiceinternalmodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)]
        )

        assert (
            service_controller.get_service_id(None, TEST_SERVICE_NAME, None)
            == TEST_SERVICE_ID
        )

        # Test the config file codepath.
        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )
        assert (
            service_controller.get_service_id(None, None, "test-config-file")
            == TEST_SERVICE_ID
        )

    def test_name_provided_multiple_matching_services(self, mock_auth_api_client):
        """If a name is provided and multiple services match it, an exception should be raised."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = ServiceinternalmodelListResponse(
            results=[Mock(id="test-service-id-1"), Mock(id="test-service-id-2")]
        )

        with pytest.raises(
            click.ClickException,
            match=f"There are multiple services with name '{TEST_SERVICE_NAME}'.",
        ):
            service_controller.get_service_id(None, TEST_SERVICE_NAME, None)

        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        with pytest.raises(
            click.ClickException,
            match=f"There are multiple services with name '{TEST_SERVICE_NAME}'.",
        ):
            service_controller.get_service_id(None, None, "test-config-file")

    def test_no_project_id(self, mock_auth_api_client):
        """If there is no project_id, the infer_project_id logic should be used."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = ServiceinternalmodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)]
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(None, TEST_SERVICE_NAME, None)
                == TEST_SERVICE_ID
            )
            service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
                project_id="mock_default_project_id", name=TEST_SERVICE_NAME,
            )

    def test_project_id_in_service_config(self, mock_auth_api_client):
        """If the project_id is passed in the service config it should be used as a filter."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = ServiceinternalmodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)]
        )

        mock_service_config = Mock()
        mock_service_config.configure_mock(
            name=TEST_SERVICE_NAME, project_id="test-project-id"
        )
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(None, None, "test-config-file")
                == TEST_SERVICE_ID
            )
            service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
                project_id="test-project-id", name=TEST_SERVICE_NAME,
            )


def test_terminate_service(mock_auth_api_client):
    """Test that v2 IDs go through v2 API, v1 through v1."""
    service_controller = ServiceController()

    service_controller.terminate("service2_abc123")
    service_controller.api_client.terminate_service_api_v2_services_v2_service_id_terminate_post.assert_called_once_with(
        "service2_abc123",
    )

    service_controller.job_controller._resolve_job_object = Mock(
        return_value=Mock(id="service_abc123"),
    )
    service_controller.terminate("service_abc123")
    service_controller.api_client.terminate_job_api_v2_decorated_ha_jobs_production_job_id_terminate_post.assert_called_once_with(
        "service_abc123",
    )


def test_rollback(mock_auth_api_client):
    """Test that v2 IDs go through v2 API, v1 is rejected."""
    service_controller = ServiceController()

    service_controller.rollback("service2_abc123")
    service_controller.api_client.rollback_service_api_v2_services_v2_service_id_rollback_post.assert_called_once_with(
        "service2_abc123",
    )

    with pytest.raises(
        click.ClickException, match="rollback is only supported for v2 services"
    ):
        service_controller.rollback("service_abc123")


def test_archive(mock_auth_api_client):
    """Test that v1 IDs go through v1 API, v2 is rejected."""
    service_controller = ServiceController()

    with pytest.raises(
        click.ClickException, match="archive is not currently supported for v2 services"
    ):
        service_controller.archive("service2_abc123")

    service_controller.job_controller._resolve_job_object = Mock(
        return_value=Mock(id="service_abc123"),
    )
    service_controller.archive("service_abc123")
    service_controller.api_client.archive_job_api_v2_decorated_ha_jobs_production_job_id_archive_post.assert_called_once_with(
        "service_abc123",
    )



from typing import Any, Dict, List, Optional

from ray.autoscaler._private.gcp.config import get_node_type
from ray.autoscaler._private.gcp.node import GCPNode
from ray.autoscaler._private.gcp.node_provider import GCPNodeProvider

from anyscale.autoscaler.node_provided_cache import NodeProviderCache


node_cache = NodeProviderCache()


class AnyscaleGCPNodeProvider(GCPNodeProvider):  # type: ignore
    def __init__(self, provider_config: Dict[str, Any], cluster_name: str) -> None:
        super().__init__(provider_config, cluster_name)

        self.provider_cache = node_cache

    def create_node(
        self, node_config: Dict[str, Any], tags: Dict[str, str], count: int
    ) -> Dict[str, Dict[str, str]]:
        """Override parent implementation.

        The logic around GCP is exactly the same, but instance_id should be
        returned, which cannot be accessed by calling the parent method.
        """
        with self.lock:
            labels = tags  # gcp uses "labels" instead of aws "tags"

            node_type = get_node_type(node_config)
            resource = self.resources[node_type]

            gcp_result = resource.create_instances(node_config, labels, count)
            instance_dict: Dict[str, Dict[str, str]] = {}
            for _, instance_id in gcp_result:
                # create_instances does not provide "PublicIpAddress"
                public_ip_address = super().external_ip(instance_id)
                instance_dict[instance_id] = {"PublicIpAddress": public_ip_address}

            return instance_dict

    def non_terminated_nodes(self, tag_filters: Dict[str, str]) -> List[str]:
        """Override parent implementation.

        The logic around GCP is exactly the same, but access to the nodes is needed to handle cache
        properly, which cannot be accessed by calling the parent method.
        """
        with self.lock:
            instances = []

            for resource in self.resources.values():
                node_instances = resource.list_instances(tag_filters)
                instances += node_instances

            # Note: All the operations use "name" as the unique instance id
            # Clear and re-populate cache.
            self.provider_cache.cleanup()
            for instance in instances:
                self.provider_cache.set_node(instance["name"], instance)
                labels = instance.get_labels()
                self.provider_cache.set_tags(instance["name"], labels)

            return [instance["name"] for instance in instances]

    def set_node_tags(self, node_id: str, node_tags: Dict[Any, Any]) -> None:
        super().set_node_tags(node_id=node_id, tags=node_tags)

        # Populate cache.
        self.provider_cache.set_tags(node_id, node_tags)

    def node_tags(self, node_id: str) -> Dict[Any, Any]:
        # Check cache first.
        if self.provider_cache.tags_exist(node_id):
            return self.provider_cache.get_tags(node_id)

        instance = super()._get_cached_node(node_id)

        # Populate cache.
        self.provider_cache.set_node(node_id, instance)
        labels = instance.get_labels()
        self.provider_cache.set_tags(instance["name"], labels)

        return labels

    def terminate_node(self, node_id: str) -> Dict[Any, Any]:
        # Delete from cache.
        self.provider_cache.delete_node_and_tags(node_id)

        result = super().terminate_node(node_id)
        return result

    def terminate_nodes(self, node_ids: List[str]) -> Optional[Dict[str, Any]]:
        # Delete from cache.
        for node_id in node_ids:
            self.provider_cache.delete_node_and_tags(node_id)

        result = super().terminate_nodes(node_ids)
        return result

    def _get_node(self, node_id: str) -> GCPNode:
        # Side effect: clear and updates cache of all nodes.
        instance = super()._get_node(node_id)

        # Populate cache.
        self.provider_cache.set_node(node_id, instance)
        labels = instance.get_labels()
        self.provider_cache.set_tags(instance["name"], labels)

        return instance

    def _get_cached_node(self, node_id: str) -> GCPNode:
        # Check cache first.
        if self.provider_cache.node_exists(node_id):
            return self.provider_cache.get_node(node_id)

        instance = super()._get_cached_node(node_id)

        # Populate cache.
        self.provider_cache.set_node(node_id, instance)
        labels = instance.get_labels()
        self.provider_cache.set_tags(instance["name"], labels)

        return instance

    def prepare_for_head_node(self, cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        * Just return cluster_config for now, unless similar credential
            issues for AWS also happen for GCP (i.e. see prepare_for_head_node() in
            frontend/cli/anyscale/autoscaler/aws/node_provider.py)
        """

        return cluster_config

    def describe_instances(self, node_ids: List[str]) -> Any:  # noqa: ARG002
        """Left describe_instances here to to be backward compatible
        (i.e. mimic AnyscaleAWSNodeProvider interface), which is AWS-specific logic
        """
        raise NotImplementedError

    def describe_gcp_instances(self, node_ids: List[str]) -> Dict[str, str]:
        """To mimic describe_instances in AnyscaleAWSNodeProvider.
        TODO: Should move the operation logic to GCPResource/GCPCompute/GCPTPU instead
        """
        with self.lock:
            instances: Dict[str, str] = {}
            for node_id in node_ids:
                resource = self._get_resource_depending_on_node_name(node_id)
                instance = (
                    resource.resource.instances()
                    .get(
                        project=self.provider_config["project_id"],
                        zone=self.provider_config["availability_zone"],
                        instance=node_id,
                    )
                    .execute()
                )

                instances[node_id] = instance

            return instances

    def start_nodes(
        self, node_ids: List[str], wait_for_operation: bool = True
    ) -> Dict[str, str]:
        """To mimic AnyscaleAWSNodeProvider
        TODO: Should move the operation logic to GCPResource/GCPCompute/GCPTPU instead
        """
        with self.lock:
            results: Dict[str, str] = {}
            for node_id in node_ids:
                resource = self._get_resource_depending_on_node_name(node_id)
                operation = (
                    resource.resource.instances()
                    .start(
                        project=self.provider_config["project_id"],
                        zone=self.provider_config["availability_zone"],
                        instance=node_id,
                    )
                    .execute()
                )

                if wait_for_operation:
                    result = resource.wait_for_operation(operation)
                else:
                    result = operation

                results[node_id] = result

            return results

    def stop_nodes(
        self, node_ids: List[str], wait_for_operation: bool = True
    ) -> Dict[str, str]:
        """To mimic AnyscaleAWSNodeProvider
        TODO: Should move the operation logic to GCPResource/GCPCompute/GCPTPU instead
        """
        with self.lock:
            results: Dict[str, str] = {}
            for node_id in node_ids:
                resource = self._get_resource_depending_on_node_name(node_id)
                operation = (
                    resource.resource.instances()
                    .stop(
                        project=self.provider_config["project_id"],
                        zone=self.provider_config["availability_zone"],
                        instance=node_id,
                    )
                    .execute()
                )

                if wait_for_operation:
                    # stop nodes takes a long time
                    result = resource.wait_for_operation(operation, max_polls=36)
                else:
                    result = operation

                results[node_id] = result

            return results



import os
from pathlib import Path
import tempfile
from types import SimpleNamespace
from typing import List
from unittest.mock import Mock, patch

import click
import pytest
import yaml

from anyscale.controllers.job_controller import LogsLogger
from anyscale.controllers.schedule_controller import (
    assert_single_parameter,
    CreateScheduleConfig,
    load_yaml_file_with_overrides,
    ScheduleConfig,
    ScheduleController,
)


CONDA_DICT = {"dependencies": ["pip", {"pip": ["pip-install-test==0.5"]}]}
PIP_LIST = ["requests==1.0.0", "pip-install-test"]
ENV_VARS_DICT = {"TEST_ENV_VAR": "test_value"}


class FakeLogger(LogsLogger):
    logs: List[str] = []

    def open_block(self, *args, **kwargs,) -> None:
        pass

    def log(self, msg: str):
        self.logs.append(msg)
        print(msg)


@pytest.fixture()
def test_directory():
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = Path(tmp_dir)
        subdir = path / "subdir"
        subdir.mkdir(parents=True)
        requirements_file = subdir / "requirements.txt"
        with requirements_file.open(mode="w") as f:
            print("\n".join(PIP_LIST), file=f)

        good_conda_file = subdir / "good_conda_env.yaml"
        with good_conda_file.open(mode="w") as f:
            yaml.dump(CONDA_DICT, f)

        bad_conda_file = subdir / "bad_conda_env.yaml"
        with bad_conda_file.open(mode="w") as f:
            print("% this is not a YAML file %", file=f)

        old_dir = os.getcwd()
        os.chdir(tmp_dir)
        yield subdir, requirements_file, good_conda_file, bad_conda_file
        os.chdir(old_dir)


@pytest.fixture()
def patch_jobs_anyscale_api_client(base_mock_anyscale_api_client: Mock):
    base_mock_anyscale_api_client.get_cluster_environment_build = Mock(
        return_value=Mock(result=Mock(status="succeeded"))
    )
    with patch.multiple(
        "anyscale.cluster_env",
        get_auth_api_client=Mock(
            return_value=Mock(anyscale_api_client=base_mock_anyscale_api_client)
        ),
    ):
        yield


s_config = {
    "name": "schedule",
    "entrypoint": "ls",
    "build_id": "build_id",
    "compute_config_id": "cc_id",
    "schedule": {"cron_expression": "ce"},
    "project_id": "prj_id",
    "runtime_env": {"pip": ["requests"]},
}


@pytest.fixture()
def schedule_yaml_file():
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = Path(tmp_dir)
        subdir = path / "subdir"
        subdir.mkdir(parents=True)

        schedule_file = subdir / "schedule.yaml"
        with schedule_file.open(mode="w") as f:
            yaml.dump(s_config, f)

        old_dir = os.getcwd()
        os.chdir(tmp_dir)
        yield schedule_file
        os.chdir(old_dir)


def test_load_yaml_file(schedule_yaml_file, mock_auth_api_client):
    config = load_yaml_file_with_overrides(schedule_yaml_file, abc=123, name=None)
    assert config == {**s_config, "abc": 123}

    config = load_yaml_file_with_overrides(schedule_yaml_file, name="config_name")
    assert config == {**s_config, "name": "config_name"}


PIP_LIST = ["requests==1.0.0", "pip-install-test"]


def test_rewrite_re(mock_auth_api_client, patch_jobs_anyscale_api_client):
    with patch.multiple(
        "anyscale.controllers.schedule_controller",
        override_runtime_env_for_local_working_dir=Mock(
            return_value={"rewritten": True}
        ),
    ):
        config = CreateScheduleConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"pip": PIP_LIST},
            schedule=ScheduleConfig(cron_expression="* * * * *"),
            project_id="test_project_id",
        )
        assert config.runtime_env == {"rewritten": True}


def test_apply_schedule(
    schedule_yaml_file, mock_auth_api_client, patch_jobs_anyscale_api_client
):
    schedule_controller = ScheduleController()
    schedule_controller.api_client.get_compute_template_api_v2_compute_templates_template_id_get = Mock(
        return_value=Mock(result=Mock(archived_at=False))
    )
    with patch.multiple(
        "anyscale.controllers.schedule_controller",
        override_runtime_env_for_local_working_dir=Mock(
            return_value={"rewritten": True}
        ),
    ):
        schedule_controller.apply(schedule_yaml_file, "override_name")
        args = schedule_controller.api_client.create_or_update_job_api_v2_experimental_cron_jobs_put.call_args[
            0
        ][
            0
        ]
        assert args.to_dict() == {
            "config": {
                "build_id": "build_id",
                "compute_config": None,
                "compute_config_id": "cc_id",
                "entrypoint": "ls",
                "max_retries": 5,
                "ray_serve_config": None,
                "runtime_env": {"rewritten": True},
                "runtime_env_config": None,
            },
            "description": None,
            "name": "override_name",
            "project_id": "prj_id",
            "schedule": {"cron_expression": "ce", "timezone": None},
        }


def test_assert_single_parameter_valid_cases(mock_auth_api_client):
    assert_single_parameter(name="name")
    assert_single_parameter(name="name", description=None, file=None)
    assert_single_parameter(name=None, description="description", file=None)
    assert_single_parameter(name=None, description=None, file="file")


def test_assert_single_parameter_invalid_cases(mock_auth_api_client):
    with pytest.raises(click.ClickException):
        assert_single_parameter()

    with pytest.raises(click.ClickException):
        assert_single_parameter(name="name", description="description")

    with pytest.raises(click.ClickException):
        assert_single_parameter(name=None, description=None)

    with pytest.raises(click.ClickException):
        assert_single_parameter(name="name", description="description", file=None)

    with pytest.raises(click.ClickException):
        assert_single_parameter(name="name", description=None, file="f")


def test_resolve_id(mock_auth_api_client):
    controller = ScheduleController()
    controller.schedule_api = Mock()
    controller.schedule_api.get.return_value = SimpleNamespace(id="returned_id")

    result = controller.resolve_file_name_or_id(id="id")
    assert result == "returned_id"


def test_resolve_id_fail(mock_auth_api_client):
    controller = ScheduleController()
    controller.schedule_api = Mock()

    def r(*args, **kwargs):
        raise Exception()

    controller.schedule_api.get = r

    with pytest.raises(click.ClickException):
        controller.resolve_file_name_or_id(id="id")


def test_resolve_name(mock_auth_api_client):
    controller = ScheduleController()
    controller.schedule_api = Mock()
    controller.schedule_api.get.return_value = SimpleNamespace(id="returned_id")
    controller.schedule_api.list.return_value = [
        SimpleNamespace(id="returned_id", name="schedule_1"),
        SimpleNamespace(id="returned_id_2", name="schedule_2"),
    ]
    result = controller.resolve_file_name_or_id(name="schedule_1")
    assert result == "returned_id"


def test_resolve_name_too_many(mock_auth_api_client):
    controller = ScheduleController()
    controller.schedule_api = Mock()
    controller.schedule_api.get.return_value = SimpleNamespace(id="returned_id")
    controller.schedule_api.list.return_value = [
        Mock(id="returned_id", name="schedule_1"),
        Mock(id="returned_id_2", name="schedule_1"),
    ]
    with pytest.raises(click.ClickException):
        controller.resolve_file_name_or_id(name="schedule_1")


def test_resolve_name_too_few(mock_auth_api_client):
    controller = ScheduleController()
    controller.schedule_api = Mock()
    controller.schedule_api.get.return_value = SimpleNamespace(id="returned_id")
    controller.schedule_api.list.return_value = []
    with pytest.raises(click.ClickException):
        controller.resolve_file_name_or_id(name="schedule_1")


def test_resolve_file(
    schedule_yaml_file, mock_auth_api_client, patch_jobs_anyscale_api_client
):
    schedule_controller = ScheduleController()
    schedule_controller.api_client.get_compute_template_api_v2_compute_templates_template_id_get = Mock(
        return_value=Mock(result=Mock(archived_at=False))
    )
    with patch.multiple(
        "anyscale.controllers.schedule_controller",
        override_runtime_env_for_local_working_dir=Mock(
            return_value={"rewritten": True}
        ),
    ):
        schedule_controller.schedule_api = Mock()
        schedule_controller.schedule_api.list.return_value = [
            SimpleNamespace(id="returned_id", name="schedule"),
        ]
        id = schedule_controller.resolve_file_name_or_id(  # noqa: A001
            schedule_config_file=schedule_yaml_file
        )
        assert id == "returned_id"


def test_pause(mock_auth_api_client):
    schedule_controller = ScheduleController()
    schedule_controller.schedule_api = Mock()
    schedule_controller.pause(id="id")
    schedule_controller.schedule_api.pause.assert_called_once()


def test_trigger(mock_auth_api_client):
    schedule_controller = ScheduleController()
    schedule_controller.schedule_api = Mock()
    schedule_controller.trigger(id="id")
    schedule_controller.schedule_api.trigger.assert_called_once()
